"""
Register Infrastructure for different Data & ML Use Cases.
"""
import json
from typing import Mapping
from datetime import datetime
from gru.utils.entity_type import EntityType
from typing import Dict


class Infrastructure:
    """
    Infrastructure baiscally contains all the details regarding the infra set up for the entire pipeline.

    Parameters
    ----------
    self: str
        Python self object
    client_id: str
        Id of the client data object
    cloud_provider:str
       Name of the cloud provider
    cluster_name: str
        Name of the cluster
    region: str
        Name of the region in which cluster is to deployed.
    subnet_ids: list
        List of subnet ids
    resource_node_group_instance_types: Mapping[str, Mapping[str, str]]
        It is is dictionary containing details about the different sizes of nodegroups.
    resource_node_group_scaling_config: Mapping[str, Mapping[str, Mapping[str, int]]]
        Corresponding to every node group it will have details about max_size, min_size ,desired_size of the node group

    admins: list
        List of all admins
    airflow_users: Mapping[str, Mapping[str, str]]
        dictionary having details about airflow users
    slack_details: Mapping[str, Mapping[str, str]]
      It will contain  host_url and host_password of slack for failure alerts and any notifications
    created_at=datetime.now()
       Time when it is created


    Returns
    -------
    None

    Notes
    -----
    Time to register the Infrastructure data  in Stuart and then will be written to DB. 

    Examples
    --------
    """

    def __init__(
        self,
        client_id: str,
        cloud_provider: str,
        cluster_name: str,
        region: str,
        subnet_ids: list,
        resource_node_group_instance_types: Mapping[str, Mapping[str, str]],
        resource_node_group_scaling_config: Mapping[
            str, Mapping[str, Mapping[str, int]]
        ],
        admins: list,
        airflow_users: Mapping[str, Mapping[str, str]],
        slack_details: Mapping[str, Mapping[str, str]],
        created_at=datetime.now(),
    ):
        self.type = EntityType.INFRASTRUCTURE.value
        self.entity_type = self.type
        self.client_id = client_id
        self.cloud_provider = cloud_provider
        self.cluster_name = cluster_name
        self.region = region
        self.subnet_ids = subnet_ids
        self.resource_node_group_instance_types = resource_node_group_instance_types
        self.resource_node_group_scaling_config = resource_node_group_scaling_config
        self.admins = admins
        self.airflow_users = airflow_users
        self.slack_details = slack_details
        self.created_at = created_at

    def to_json(self) -> Dict:
        return {
            "type": self.type,
            "client_id": self.client_id,
            "cloud_provider": self.cloud_provider,
            "cluster_name": self.cluster_name,
            "region": self.region,
            "subnet_ids": self.subnet_ids,
            "resource_node_group_instance_types": json.dumps(
                self.resource_node_group_instance_types
            ),
            "resource_node_group_scaling_config": json.dumps(
                self.resource_node_group_scaling_config
            ),
            "admins": self.admins,
            "airflow_users": json.dumps(self.airflow_users),
            "slack_details": json.dumps(self.slack_details),
            "created_at": str(self.created_at),
        }



"""
Pre-Processing Transforms are used to schedule pipelines by providing a custom business logic. This helps create additional Datasources on top of which features can be defined.
testind in production
Functions:
- register
- deploy
"""

from datetime import datetime
from typing import Optional, Dict
from gru.schema.registrable import Registrable
from gru.utils.entity_type import EntityType
from gru.features.processing_engine_configs import BatchProcessingEngineConfigs
from gru.utils.processing_engine import ProcessingEngine

class PreProcessingTransform(Registrable):
    """
    A preprocessing transform represents a custom pipeline used to create intermediate datasources

    Parameters
    ----------
    name: str
        name of the preprocess transform job
    description: str
        description of the preprocess transform job
    data_source_names: list
        list of names of the data sources registered in our platform
    data_source_lookback_config: Dict
        a dict of keys as feature name and value as the lookback window for that particular feature
    logic: str
        SQL logic of the preprocess tranform job
    schedule: str
        schedule of the preprocess transform job
    output_schema: Schema
        schema of the output data persisted in final blob storage
    owners: list
        list of the owners of Preprocess transform job
    active: Optional[str]
        if True, PPT job will start running startdate onwards. And if False, PPT job will be registered and deployed but execution won't begin
    stage: Optional[str]
        stage is the environment e.g. dev or prod
    timeout: Optional[str]
        if job takes more than timeout value, then users will be notified over slack.
    cluster: Optional[str]
        Kubernetes cluster on which the job will run, default values set by Yugen
    k8s_namespace: Optional[str]
        namespace of Kubernetes cluster, default values set by Yugen
    transform_start_time: Optional[datetime]
        start time of the PPT job, default values set by Yugen to datetime.now()

    Returns
    -------
    None

    Methods
    -------
    register(): takes self as an argument and register the Preprocess transform job to the database.
    deploy(): takes self as an argument and deploys the Preprocess transform job. PPT job starts execution as per the PPT start date.

    Notes
    -----

    See Also
    --------

    Examples
    --------
    >>>     ppt_obj = PreProcessingTransform(
    ...         name = "test_transform_demo_v9_sm",
    ...         description = "test preprocess transform",
    ...         data_source_names = ["survey_tracking3", "feedback3"],
    ...         data_source_lookback_config = {
    ...         "survey_tracking3": "1d",
    ...         "feedback3": "2d",
    ...         },
    ...         logic = sql_logic,
    ...         schedule = "0 0 * * *",
    ...         output_schema = schema_obj,
    ...         owners = ["manik.malhotra@yugen.ai"],
    ...         active = True,
    ...         stage = "dev",
    ...         timeout = "5m",
    ...         cluster = "demo_cluster",
    ...         k8s_namespace = "airflow",
    ...         transform_start_time = date(2022, 8, 28),
    ...         )
    """

    def __init__(
        self,
        transform_name: str,
        description: str,
        data_source_names: list,
        data_source_lookback_config: Dict,
        logic: str,
        schedule: str,
        output_schema: object,
        owners: list,
        staging_sink: list[str],
        active: bool = False,
        timeout: str = "5m",
        processing_engine: Optional[ProcessingEngine] = ProcessingEngine.PYSPARK_K8S,
        processing_engine_configs: Optional[BatchProcessingEngineConfigs] = None,
        k8s_configs: Dict = {},
        transform_start_time: Optional[datetime] = datetime.now(),
        online_sink: Optional[list[str]] = None,
    ):
        self.sink = {"staging": staging_sink, "online": online_sink}
        self.type = EntityType.PRE_PROCESSING_TRANSFORM
        self.entity_type = self.type.value
        self.transform_name = transform_name
        self.description = description
        self.data_source_names = data_source_names
        self.data_source_lookback_config = data_source_lookback_config
        self.logic = logic
        self.schedule = schedule
        self.output_schema = output_schema
        self.owners = owners
        self.active = active
        self.processing_engine = processing_engine
        self.processing_engine_configs = (
            processing_engine_configs or BatchProcessingEngineConfigs()
        )
        self.k8s_configs = k8s_configs
        self.timeout = timeout
        self.airflow_config = {
            "active": self.active,
            "schedule": self.schedule,
            "timeout": self.timeout,
        }
        self.transform_start_time = transform_start_time
        self.creation_time = datetime.now()

    def to_json(self) -> Dict:
        """
        Returns a json-formatted object representing this entity.
        """
        return {
            "transform_name": self.transform_name,
            "description": self.description,
            "data_source_names": self.data_source_names,
            "data_source_lookbacks": self.data_source_lookback_config,
            "logic": self.logic,
            "output_schema": self.output_schema.schema,
            "owners": self.owners,
            "processing_engine": self.processing_engine.value,
            "processing_engine_configs": self.processing_engine_configs.to_json(),
            "k8s_configs": self.k8s_configs,
            "transform_start_time": str(self.transform_start_time),
            "creation_time": str(self.creation_time),
            "airflow_configs": self.airflow_config,
            "type": self.type.value,
            "sink": self.sink,
        }

    def to_register_json(self) -> Dict:
        return self.to_json()



from datetime import datetime
from typing import Any, Dict, List, Literal, Optional
from pydantic import BaseModel
from gru.models.common.hpa_config import HPAConfig

class WorkflowListQueryParams(BaseModel):
    """Query parameters for listing workflows"""    
    is_active: Optional[bool] = None

class WorkflowDeployment(BaseModel):
    """Schema for workflow deployment details"""
    cluster_name: str
    deployment_status: Literal['PENDING', 'DEPLOYED', 'FAILED', 'TERMINATED']
    last_synced_at: Optional[datetime]

class Workflow(BaseModel):
    """Schema for workflow response"""
    workflow_name: str
    description: Optional[str]
    status: Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']
    deployments: List[WorkflowDeployment]
    created_at: datetime
    updated_at: datetime

class RuleDetails(BaseModel):
    """Schema for rule details within a workflow"""
    rule_name: str
    rule_definition: Dict[str, Any]
    stage: Literal['REGISTERED', 'BACKTESTING', 'STAGING', 'PRODEXPERIMENT', 'LIVE']
    status: Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']
    created_at: datetime
    updated_at: datetime

class WorkflowDetails(BaseModel):
    """Schema for detailed workflow response including rules"""
    workflow_name: str
    description: Optional[str]
    status: Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']
    rules: List[RuleDetails]
    deployments: List[WorkflowDeployment]
    created_at: datetime
    updated_at: datetime

class SubRule(BaseModel):
    field: str
    operator: Literal['<', '>', '==', '!=', '>=', '<=']
    value: Optional[Any] = None 
    storage_key: Optional[str] = None
    storage_field: Optional[str] = None

class CreateRuleRequest(BaseModel):
    rule_name: str
    operator: Literal['AND', 'OR']
    sub_rules: List[SubRule]

class CreateWorkflowWithRuleRequest(BaseModel):
    """Request model for creating a new workflow with initial rules"""
    workflow_name: str
    description: Optional[str]
    initial_rules: List[CreateRuleRequest] 

class UpdateWorkflowStatusRequest(BaseModel):
    """Schema for workflow status update request"""
    new_status: Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']

    @classmethod
    def from_str(cls, status: str) -> 'UpdateWorkflowStatusRequest':
        """
        Create request from string status, with validation
        """
        status_upper = status.upper()
        if status_upper not in ('ACTIVE', 'INACTIVE', 'DEPRECATED'):
            raise ValueError(
                "Status must be one of: ACTIVE, INACTIVE, DEPRECATED"
            )
        return cls(new_status=status_upper) 

class UpdateWorkflowStatusResponse(BaseModel):
    """Response model for workflow status update"""
    workflow_name: str
    description: Optional[str]
    status: Literal['ACTIVE', 'INACTIVE', 'DEPRECATED']
    created_at: datetime
    updated_at: datetime

class DeployWorkflowRequest(BaseModel):
    """Request model for workflow deployment"""
    cluster_name: str
    environment: Literal["BACKTESTING", "STAGING", "PRODEXPERIMENT", "LIVE"]
    namespace: str
    env_vars: Dict[str, str] = {}
    hpa_configs: Optional[HPAConfig] = None

    class Config:
        allow_population_by_field_name = True

class DeployWorkflowResponse(BaseModel):
    """Response model for workflow deployment"""
    
    deployment_strategy: str
    deployment_id: str
    workflow_name: str
    environment: str
    argocd_app_name: str
    namespace: str
    chart: Optional[str] = None
    version: Optional[str] = None


import json
import logging
import os
from typing import Mapping

from gru.agents.clients.rabbitmq import RabbitMQPublisher
from gru.agents.utils import constants

class RabbitMQLogHandler(logging.Handler):

    ALLOWED_EXTRA_FIELDS = ["app", "correlation_id", "tenant_id"]

    def __init__(self, publisher: RabbitMQPublisher, level=logging.NOTSET):
        self.publisher = publisher
        super(RabbitMQLogHandler, self).__init__(level=level)
        self.queue = os.getenv("CANSO_INCOMING_QUEUE")

    def emit(self, record: logging.LogRecord):
        try:
            log_message = self._format(record)
            self.publisher.publish(self.queue, self.queue, log_message)
        except Exception as e:
            print(f"Exception while sending log to rabbitmq: {e}")

    def _format(self, record: logging.LogRecord) -> str:
        extra_fields = {
            k: v for k, v in record.__dict__.items() if k in self.ALLOWED_EXTRA_FIELDS
        }
        message = {
            "message_type": "log",
            "message_body": {
                "level_name": record.levelname,
                "log_message": record.getMessage(),
                "extra": extra_fields,
            },
        }

        return json.dumps(message)

    def close(self):
        self.publisher.stop()
        super().close()

def get_log_fields(correlation_id: str) -> Mapping[str, any]:
    return {
        "app": constants.APP_NAME,
        "tenant_id": constants.TENANT_ID,
        "correlation_id": correlation_id,
    }



"""
To update existing raw and derived features. Supported values - 
    - "owners"
    - "description"
    - "start_time"
    - "online"
    - "offline"
    - "active"
"""

from typing import Dict
from datetime import datetime


class UpdateFeature:
    """
    Update a feature 

    Parameters
    ----------
    feature_name: str
        feature name to be updated
    property: str
        name of the property to be updated
    new_val: str
        new value for the feature property
    
    Returns
    -------
    None

    Notes
    -----

    Examples
    --------
    >>> obj = UpdateFeature(
    ...         name= 'raw_feature_test_v2',
    ...         property= 'active',
    ...         new_val= False,
    ... )
    ... _
    ... yugen_client.update(update_obj)
    """

    def __init__(self, name, property, new_val):
        self.name = name
        self.property = property
        self.new_val = new_val

        if isinstance(self.new_val, datetime):
            self.new_val = str(self.new_val)


    def to_json(self) -> Dict:
        """
        Convert the UpdateFeature object to a JSON-compatible dictionary.
        """
        return {
            "name": self.name,
            "property": self.property,
            "new_value": self.new_val,
        }



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from typing import Optional
from gru.utils.commons import dt_to_str
from gru.utils.commons import is_duration_valid_for_dry_run
from gru.utils.config_reader import ConfigReader
from gru.utils.entity_type import EntityType


class DeploymentRequest(ABC):
    """
    Abstract base class representing a deployment strategy.
    """

    def __init__(
        self,
        workflow_name: str,
        entity_type: EntityType,
        dry_run: bool = False,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        config=None,
    ):
        self.dry_run = dry_run
        self.start_date = start_date
        self.end_date = end_date
        self.config = config
        self.workflow_name = workflow_name
        self.entity_type = entity_type

    @abstractmethod
    def create_deployment_options(self) -> dict:
        """
        Creates a dictionary containing options specific to the deployment strategy.

        Returns:
            dict: The deployment options dictionary.
        """

        pass

    @abstractmethod
    def is_entity_deployable(self):
        pass


class AirflowDeploymentRequest(DeploymentRequest):
    """
    Deployment strategy using Airflow workflows.
    """

    deployable_entities = {EntityType.RAW_FEATURE, 
                           EntityType.DERIVED_FEATURE, 
                           EntityType.PRE_PROCESSING_TRANSFORM,
                           EntityType.STREAMING_FEATURE}

    def __init__(
        self,
        workflow_name: str,
        entity_type: EntityType,
        config: ConfigReader,
        dry_run: bool = False,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ):
        """
        Args:
            workflow_name (str): Name of the Airflow workflow to trigger.
            **kwargs: Additional keyword arguments passed to the parent class constructor.
        """
        super().__init__(
            workflow_name=workflow_name,
            entity_type=entity_type,
            config=config,
            dry_run=dry_run,
            start_date=start_date,
            end_date=end_date,
        )

    def create_deployment_options(self) -> dict:
        """
        Returns options specific to Airflow deployment.

        Returns:
            dict: A dictionary containing the Airflow workflow name and optional start/end dates.
        """

        url_path = self.config.deploy_paths[self.entity_type.value]
        json_req = {
            "name": self.workflow_name,
            "type": self.entity_type.value,
            "airflow_dags_backup_path": (
                self.config.airflow_dags_backup_path
                if self.config.airflow_dags_backup_path
                else None
            ),
        }

        if self.dry_run:
            self._validate_dry_run()
            json_req.update({
                "is_dry_run": True,
                "start_date": str(dt_to_str(self.start_date)),
                "end_date": str(dt_to_str(self.end_date))
            })

        return {"url_path": url_path, "json": json_req}
    
    def _validate_dry_run(self):
        max_days = self.config.dry_run["max_dry_run_duration_days"]
        if not is_duration_valid_for_dry_run(self.start_date, self.end_date, max_days):
            raise ValueError(
                f"""max_dry_run_duration_days is set to {max_days}. Adjust your start and end date accordingly."""
            )
        if self.end_date is None:
            self.end_date = self.start_date + timedelta(
                days=self.config.dry_run["default_dry_run_duration"]
            )
    
    def is_entity_deployable(self):
        if self.entity_type not in self.deployable_entities:
            return False
        
        return True



from abc import ABC, abstractmethod
from typing import List

class EmbeddingGenerator(ABC):

    @property
    @abstractmethod
    def dimension(self) -> int:
        pass

    @abstractmethod
    async def generate(self, text: str) -> List[float]:
        pass

    @abstractmethod
    async def generate_batch(self, texts: List[str]) -> List[List[float]]:
        pass


from datetime import datetime
from typing import Dict, Optional
from urllib.parse import urljoin
import requests
from gru.utils.entity_type import EntityType


class APIRequestHandler:
    """
    Handles sending requests to the Yugen remote API.
    """

    def __init__(self, auth_token, config=None, params = None, base_url:str = None):
        self.params = params or {}
        self.config = config
        self.base_url = base_url if base_url is not None else self.config.stuart_api_endpoint
        #TODO: 22-Mar-24    Sandeep Mishra     add authentication header once jwt token implemented accross APIs in Stuart
        self.headers = {
            "Content-type": "application/json",
            "Accept": "text/plain",
            "Authorization": f"Bearer {auth_token}"
        }

    def set_headers(self, headers: dict):
        """
        Updates the request headers with the provided dictionary.
        """
        self.headers.update(headers)
    
    def set_base_url(self, base_url: str):
        """
        Updates base url of the Request handler
        """
        self.base_url=base_url

    def set_authorization(self, token: str):
        """
        Sets the authorization token in the headers dictionary.

        Args:
            token (str): The authorization (i.e. jwt) token to be set.
        """

        self.headers["Authorization"] = "Bearer {}".format(token)
    
    def set_correlation_id(self, correlation_id:str):
        self.headers["correlation-id"] = correlation_id

    def set_params(self, params: dict):
        """
        Sets the request parameters for the next request.
        """
        self.params = params

    def send_request(
        self, method: str, url_path: str, json_data: Optional[dict] = None
    ) -> requests.Response:
        """
        Sends an HTTP request to the Yugen API using the specified method and path.

        Args:
            method (str): The HTTP method (e.g., 'POST', 'GET', 'PATCH', etc.).
            url_path (str): The path relative to the base URL.
            json_data (Optional[dict]): The JSON data to be sent in the request body.

        Returns:
            dict: The response from the API as a dictionary.

        Raises:
            requests.exceptions.RequestException: If there's an issue with the API request.
            ValueError: If a required parameter is missing.
        """

        if not method:
            raise ValueError("Method argument is required")
        if not url_path:
            raise ValueError("url_path argument is required")

        api_url = urljoin(self.base_url, url_path)

        try:
            response = requests.request(
                method=method.upper(), url=api_url, json=json_data, headers=self.headers, params=self.params
            )
            return response
        except requests.exceptions.RequestException as e:
            raise e
    
    def create_register_json(self, register_object) -> Dict:
        """
        Create JSON payload for registration.
        """
        entity_type = register_object.entity_type
        url_path = self.config.register_paths[entity_type]
        return url_path, register_object.to_json()

    def create_deploy_json(self, name: str, entity_type: EntityType,
                            dry_run: Optional[bool]= False, 
                            start_date: Optional[datetime] = None,
                            end_date: Optional[datetime] = None) -> Dict:
        """
        Create JSON payload for deployment.
        """
        url_path =  self.config.deploy_paths[entity_type.value],
        json_req = { "name": name,
                    "type": entity_type.value,
                    "airflow_dags_backup_path": self.config.airflow_dags_backup_path if self.config.airflow_dags_backup_path else None
                }
        
        if dry_run : 
            json_req.update("start_date", start_date)
            json_req.update("end_date", end_date)

        return url_path, json_req



"""
Data Sources are an abstraction on top of datasets that enable

 - A standardized way of declaring raw data on top of which features & pre-processing tables are calculated
 - A simplified experience while referencing in defining features - Data Scientists can focus on the feature logic without having to worry about the underlying data, DB connections, accesses etc.
 - Reusability while defining and materializing features/pre-processing transforms
 - Improved understanding of raw data

Functions: 
    - return_col_details
    - return_schema
    - check_argument_type
    - register_datasource
"""
from abc import abstractmethod
from datetime import datetime
import json
from typing import Any, Dict, List, Optional, Union

from gru.schema.registrable import Registrable
from gru.sources.file_types import CSVType, ParquetType
from gru.utils.data_types import DataType
from gru.utils.entity_type import EntityType


class Column:
    """
    Define the column of data source

    Parameters
    ----------
    self:
        python self object
    name:
        name of the column
    type:
        data type of the column
    nullable:
        If set to True, then the column can have Null value. If False, then the column can't be Null
    description:
        Description of the column

    Notes
    -----

    Examples
    --------
    >>> schema_obj = Schema(
    ...     user_schema=[
    ...         Column("user_id", DataTypes.STRING, True, "Unique User Identifier"),
    ...         Column("offer_id", DataTypes.STRING, True, "Offer viewed by the User"),
    ...         Column("clicked_at", DataTypes.TIMESTAMP, True, "Time of offer click"),
    ...         Column("viewed_at", DataTypes.STRING, True, "Time of Offer View"),
    ...         Column(
    ...             "view_to_click_time_mins", DataTypes.FLOAT, True, "View to Click time in minutes"
    ...         ),
    ...     ]
    ... )
    """

    def __init__(self, name, type: DataType, nullable=True, description=""):
        """
        :param name: Name of the column
        :param type: Datatype of the column
        :param nullable: Nullable or not
        :param description: Description of the column
        """
        self.name = name
        self.type = type.value
        self.nullable = nullable
        self.description = description
        self.col_details = [self.name, self.type, self.nullable, self.description]
        self.return_col_details()

    def return_col_details(self):
        return self.col_details


class Schema:
    """
    Define the schema of the data source

    Parameters
    ----------
    self:
        python self object
    user_schema: List[Column]
        user_schema is the list of columns present in the datasource.

    Notes
    -----

    Examples
    --------
    >>> schema_obj = Schema(
    ...     user_schema=[
    ...         Column("user_id", DataTypes.STRING, True, "Unique User Identifier"),
    ...         Column("offer_id", DataTypes.STRING, True, "Offer viewed by the User"),
    ...         Column("clicked_at", DataTypes.TIMESTAMP, True, "Time of offer click"),
    ...         Column("viewed_at", DataTypes.STRING, True, "Time of Offer View"),
    ...         Column(
    ...             "view_to_click_time_mins", DataTypes.FLOAT, True, "View to Click time in minutes"
    ...         ),
    ...     ]
    ... )
    """

    def __init__(self, user_schema: List[Column]):
        self.user_schema = user_schema

        self.pyspark_schema = self.convert_to_pyspark_schema()
        self.bq_schema = self.convert_to_bq_schema()

        self.return_schema(self.pyspark_schema, self.bq_schema)

    def convert_to_pyspark_schema(self):
        """Method to convert column details list to pyspark like schema and append to the schema dictionary"""
        self.schema = {}
        col_schema = []
        for column_obj in self.user_schema:
            col = column_obj.col_details
            col_schema.append(
                {
                    "metadata": {},
                    "name": col[0],
                    "nullable": col[2],
                    "type": col[1],
                }
            )
        self.schema = {"fields": col_schema, "type": "struct"}
        return self.schema

    def convert_to_bq_schema(self):
        """Method to convert column details list to BQ like schema and append to the schema dictionary"""
        self.schema = []
        for column_obj in self.user_schema:
            col = column_obj.col_details
            self.schema.append(
                {
                    "name": col[0],
                    "nullable": col[2],
                    "type": col[1],
                    "description": col[3],
                }
            )
        return self.schema

    def return_schema(self, pyspark_schema, bq_schema):
        self.schema = {
            "pyspark_schema": pyspark_schema,
            "bq_schema": bq_schema,
        }
        return str(self.schema)


class DataSource(Registrable):
    """
    A generic class for any data source that will act as the parent class for all the data sources to be supported in future.

    """

    def __init__(
        self,
        name: str,
        description: str,
        family: str,
        processing_paradigm: str,
        additional_read_configs: Optional[Dict[str, Any]] = {},
        cloud_provider: str = None,
        file_type: Union[CSVType, ParquetType] = None,
        physical_uri: Dict[str, Any] = None,
        schema: Schema = None,
        type: str = None,
        metadata: Optional[Dict[str, str]] = {},
        event_timestamp_field: str = None,
        event_timestamp_format: str = None,
        owners: Union[List[str], str] = [],
        created_at=datetime.now(),
    ):
        self.name = name
        self.family = family
        self.description = description
        self.processing_paradigm = processing_paradigm
        self.file_type = file_type
        self.entity_type = EntityType.DATA_SOURCE.value
        self.physical_uri = physical_uri
        self.additional_read_configs = additional_read_configs
        self.schema = schema
        self.type = type
        self.cloud_provider = cloud_provider
        self.metadata = metadata
        self.event_timestamp_field = event_timestamp_field
        self.event_timestamp_format = event_timestamp_format
        self.owners = owners
        self.created_at = created_at

    @abstractmethod
    def to_json(self) -> Dict:
        """
        Returns a json-formatted object representing this entity.
        """
        pass

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2, sort_keys=True)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.to_json() == other.to_json()
    
    def to_register_json(self) -> Dict:
        return self.to_json()



from enum import Enum


class ElasticCacheRedisClusterMode(Enum):
    """
    Redis Cluster Mode in Elasticache (AWS)
    """

    CLUSTER_ENABLED = "cluster_mode_enabled"
    CLUSTER_DISABLED = "cluster_mode_not_enabled"



from gru.agents.tools.core.code_generator.sql import SQLCodeGenerator
from gru.agents.tools.core.query_analyzer.nl_query import NLQueryAnalyzer
from gru.agents.tools.core.code_generator.models import (
    QueryIntent, RetrievalResult
)
from gru.agents.tools.core.llm_client.base import LLMClient
from gru.agents.tools.core.context_retriever.base import ContextRetriever
from gru.agents.tools.core.embeddings.base import EmbeddingGenerator
from gru.agents.prompts.text_to_sql import (
    QUERY_ANALYZER_USER_PROMPT_TEMPLATE,
    QUERY_ANALYZER_SYSTEM_PROMPT,
)
from typing import Optional, List, Dict, Any


class TextToSQLService:
    def __init__(
        self,
        llm_client: LLMClient,
        context_retriever: ContextRetriever,
        embedding_generator: EmbeddingGenerator,
        code_generator: Optional[SQLCodeGenerator] = None,
        query_analyzer: Optional[NLQueryAnalyzer] = None,
        system_prompt: Optional[str] = None,
        user_prompt_template: Optional[str] = None,
    ):
        self.llm_client = llm_client
        self.context_retriever = context_retriever
        self.embedding_generator = embedding_generator
        self.code_generator = code_generator or SQLCodeGenerator(llm_client)
        self.query_analyzer = query_analyzer
        self.system_prompt = system_prompt or QUERY_ANALYZER_SYSTEM_PROMPT
        self.user_prompt_template = (
            user_prompt_template or QUERY_ANALYZER_USER_PROMPT_TEMPLATE
        )

    async def convert_to_sql(
        self, query: str, table_info: Optional[List[Dict[str, Any]]] = None
    ) -> str:
        try:
            if table_info:
                await self.context_retriever.initialize_store(table_info)
            intent = await self._analyze_query(query)

            # Generate embedding for user query using the embedding generator
            query_embedding = await self.embedding_generator.generate(
                query
            )

            context = await self.context_retriever.retrieve_context(
                query_embedding, top_k=5
            )

            sql = await self._generate_sql(query, intent, context)

            # await self.context_retriever.store_conversation(query, sql)

            return sql.sql_query

        except Exception as e:
            return f"Error converting text to SQL: {str(e)}"

    async def _analyze_query(self, query: str) -> QueryIntent:
        user_prompt = self.user_prompt_template.format(query=query)
        response = await self.llm_client.generate(
            self.system_prompt, user_prompt
        )

        lines = response.split("\n")
        entities = (
            lines[0]
            .replace("Entities:", "")
            .strip()
            .strip("[]")
            .split(",")
        )

        domains = (
            lines[1]
            .replace("Domains:", "")
            .strip()
            .strip("[]")
            .split(",")
        )

        return QueryIntent(
            entities=[e.strip() for e in entities if e.strip()],
            domains=[d.strip() for d in domains if d.strip()],
        )

    async def _generate_sql(
        self, query: str, intent: QueryIntent, context: RetrievalResult
    ) -> str:
        return await self.code_generator.generate_code(query, intent, context)



from typing import Dict, List
from gru.features.aggregations import SlidingWindowAggregation, WindowAggregation, SessionWindowAggregation
from gru.features.transformations import BuiltInTransformation, LambdaTransformation


class FeatureLogic:
    """
    Create a FeatureLogic object for the feature definition

    Parameters
    ----------
    self: str
        self object of the class is the first argument
    field: str
        field on which raw feature will be calculated
    aggregation_fn: str
        aggregation function to be performed on the field
    time_window: str
        time window of the data to be fetched from the data source
    groupby_keys: str
        select fields based on which feature will be calculated
    timestamp_field: str
        timestamp field present in the data source for feature calculation
    timestamp_format: str
        format of the timestamp field
    filter: str
        filter expression that will be applied to rows in the data before performing aggregations. This should be a string and a valid Spark SQL Expression. This is similar to PySpark filter operation, See more details in the official documentation: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.filter.html
    name: str
        name of the feature logic, default set to ""
    is_feature_logic_custom: bool
        Flag to decide weather the logic is custom or predefined, Default value is True .

    Returns
    -------
    None

    Notes
    -----

    Examples
    --------
    >>>     feature_logic=FeatureLogic(
    ...         name="total_players",
    ...         field=["CASE WHEN cpi> 0.5 THEN 10 ELSE 0 END"],
    ...         aggregation_fn="SUM",
    ...         time_window= "2d",
    ...         groupby_keys=["project_id", "ad_id"],
    ...         timestamp_field="etl_day",
    ...         timestamp_format = "%Y-%m-%d %H:%M"
    ...     )
    """

    def __init__(
        self,
        name: str = "",
        field: List = "",
        transform: object = None,
        time_window: str = "",
        groupby_keys: List = [],
        timestamp_field: str = "",
        timestamp_format: str = "",
        filter: str = None,
    ):
        self.name = name
        self.filter = filter
        self.field = field

        if isinstance(transform, WindowAggregation) or isinstance(
            transform, SlidingWindowAggregation
        ) or isinstance(
            transform, SessionWindowAggregation
        ):
            self.transform = transform.transform_args
        elif isinstance(transform, BuiltInTransformation):
            self.transform = transform.to_dict()
        elif isinstance(transform,LambdaTransformation):
            self.transform=transform.to_dict()

        self.time_window = time_window
        self.groupby_keys = groupby_keys
        self.timestamp_field = timestamp_field
        self.timestamp_format = timestamp_format
        self.is_feature_logic_custom = False

    def to_json(self) -> Dict:
        """
        Returns a json-formatted object representing this entity.
        """
        return {
            "name": self.name,
            "transform": self.transform,
            "time_window": self.time_window,
            "groupby_keys": self.groupby_keys,
            "field": self.field,
            "filter": self.filter,
            "timestamp_field": self.timestamp_field,
            "timestamp_format": self.timestamp_format,
            "is_feature_logic_custom": self.is_feature_logic_custom
        }



"""
Specify FileTypes and other options that are necessary to load data from DataSources
"""
from typing import Optional


class CSVType:
    """
    Represents a CSV File/Object

    Parameters
    ----------
    #TODO - Defaults need to be corrected
    sep: Optional[str]
        delimiter; by default ','
    inferSchema: Optional[bool]
        by default set to False
    header: Optional[bool]
        if True, will read the first row of csv file as column names. by default set to False
    quotes: Optional[str]
        When you have a column with a delimiter in the data that is used to split the columns, use quotes option to specify the quote character, by default it is ” and delimiters inside quotes are ignored. Using this option you can set any quote character.
    nullValue: Optional[str]
        Using nullValue option you can specify the string in a CSV to consider as null. For example, if you want to set "manik" in name column to null
        |-- age: long (nullable = true) # 26
        |-- name: string (nullable = true) # manik
        +---+----+
        |age|name|
        +---+----+
        | 26|null|
        +---+----+
    nanValue: Optional[str]
        Using nanValue option you can specify the number in a CSV to consider as nan. For example, if you want to set 23.4 in name column to nan
    dateFormat: Optional[str]
        by default "yyyy-MM-dd HH:mm:ss"
    timestampFormat: Optional[str]
        by default "yyyy-MM-ddTHH:mm:ss.%f"
    encoding: Optional[str]
        by default set to 'utf-8' encoding

    Notes
    -----
    See Default values of Options - https://spark.apache.org/docs/latest/sql-data-sources-csv.html#data-source-option

    Examples
    --------
    >>> data_source_obj = S3DataSource(
    ...     data_source_name="my_data_source_11",
    ...     bucket="market-intelligence-platform",
    ...     base_key="sales/raw_data",
    ...     varying_key_suffix_format="d=%Y-%m-%d/h=%H%M",
    ...     varying_key_suffix_freq="15min",
    ...     time_offset=5*60,   
    ...     description="random desc of data source",
    ...     owners=["xyz"],
    ...     created_at=datetime(2022, 12, 20, 00, 00, 00),
    ...     file_type=CSVType(sep='|', header=True),
    ...     schema=schema_obj,
    ...     event_timestamp_field="viewed_at",
    ...     event_timestamp_format="%Y-%m-%d %H:%M",
    ... )
    """

    def __init__(
        self,
        sep: Optional[str] = None,
        inferSchema: Optional[bool] = None,
        header: Optional[bool] = None,
        quote: Optional[str] = None,
        nullValue: Optional[str] = None,
        nanValue: Optional[str] = None,
        dateFormat: Optional[str] = None,
        timestampFormat: Optional[str] = None,
        encoding: Optional[str] = None,
    ):
        """
        :param sep: delimiter; by default ','
        :param inferSchema: by default set to False
        :param header: if True, will read the first row of csv file as column names. by default set to False
        :param nullValue: Default `None`
        :param nanValue: Default `None`
        :param dateFormat: by default "yyyy-MM-dd HH:mm:ss"
        :param timestampFormat: by default "yyyy-MM-ddTHH:mm:ss.%f"
        :param encoding: Default `None`
        
        See Default values of Options in Spark - https://spark.apache.org/docs/latest/sql-data-sources-csv.html#data-source-option
        """

        self.sep = sep
        self.inferSchema = inferSchema
        self.header = header
        self.quote = quote
        self.nullValue = nullValue
        self.nanValue = nanValue
        self.dateFormat = dateFormat
        self.timestampFormat = timestampFormat
        self.encoding = encoding
        self.type = "CSV"

        self.options = {
            "type": self.type,
            "delimiter": self.sep,
            "inferSchema": self.inferSchema,
            "header": self.header,
            "quotes": self.quote,
            "nullValue": self.nullValue,
            "nanValue": self.nanValue,
            "dateFormat": self.dateFormat,
            "timestampFormat": self.timestampFormat,
            "encoding": self.encoding,
        }

        self.options = {k: v for k, v in self.options.items() if v is not None}


class ParquetType:
    """
    Represents a Parquet File/Object

    Parameters
    ----------
    mergeSchema: Optional[bool]
        if True, will merge schema of multiple parquet files while doing a Union, by default 'False'
    
    Notes
    -----

    Examples
    --------
    >>> data_source_obj = S3DataSource(
    ...     data_source_name="my_data_source_11",
    ...     bucket="market-intelligence-platform",
    ...     base_key="sales/raw_data",
    ...     varying_key_suffix_format="d=%Y-%m-%d/h=%H%M",
    ...     varying_key_suffix_freq="15min",
    ...     time_offset=5*60,   
    ...     description="random desc of data source",
    ...     owners=["xyz"],
    ...     created_at=datetime(2022, 12, 20, 00, 00, 00),
    ...     file_type=ParquetType(mergeSchema=True),
    ...     schema=schema_obj,
    ...     event_timestamp_field="viewed_at",
    ...     event_timestamp_format="%Y-%m-%d %H:%M",
    ... )
    """

    def __init__(self, mergeSchema: Optional[bool] = False):
        """
        :param mergeSchema: if True, will merge schema of multiple parquet files while doing a Union, by default 'False'
        """
        self.mergeSchema = mergeSchema
        self.type = "PARQUET"
        self.options = {"type": self.type, "mergeSchema": self.mergeSchema}



"""
Module consisting of classes and functions for Datatypes supported by our platform. Currently all Apache Pyspark SQL datatypes are supported.
Reference: https://sparkbyexamples.com/pyspark/pyspark-sql-types-datatype-with-examples/
"""



from enum import Enum, unique

@unique
class DataType(Enum):
    """
    Enum-like class defining data types supported by our Platform
    """

    INT = "integer"
    STRING = "string"
    TIMESTAMP = "timestamp"
    FLOAT = "float"
    DATETIME = "datetime"
    LIST = "list"
    LIST_INT = "list[integer]"
    LIST_FLOAT = "list[float]"
    LIST_STRING = "list[string]"

    @classmethod
    def get_key(cls, value):
        """
        Get the key (name) of a data type based on its value.

        Args:
            value (str): The value of the data type.

        Returns:
            str: The key (name) of the data type, or None if not found.
        """
        for item in cls:
            if item.value == value:
                return item.name
        return None



"""
The `YugenClient` class serves as a client for registering, deploying, and simulating objects using a remote API.
It provides functionalities to interact with a backend system to manage various types of feature objects.
"""

from datetime import datetime
import json
import os
from typing import Optional, Type, Union, Dict
import requests
from gru.agents.apis import AGENT_CONFIG_FILE_NAME, deploy_agent, prompt_agent, read_prompt_file, register_agent
from gru.features.update_feature import UpdateFeature
from gru.ai_agents.agent_deployment import AIAgent
from gru.ai_agents.send_rabbitmq_message import AIAgentMessage
from gru.preprocess.update_preprocessing_transform import UpdatePreprocessingTransform
from gru.schema.api_request_handler import APIRequestHandler
from gru.schema.api_response_handler import APIResponseHandler, ApiError
from gru.schema.registrable import Registrable
from gru.utils.config_reader import ConfigReader
from urllib.parse import urljoin
from gru.utils.deployment_workflow import Workflow, get_deployment_request_class
from gru.utils.entity_type import EntityType
from gru.auth.generate_access_token import GenerateAccessToken
from gru.components.apis import setup as component_setup


import logging

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class YugenClient:
    """
    A client for registering objects with a remote API.

    Attributes:
        access_token (str): The access token for authentication.
        config (ConfigReader): An instance of the configuration reader.
    """

    def __init__(
        self, auth_token: Optional[str] = None, config_path="./gru/config.yaml"
    ):
        """
        Initialize the YugenClient.

        Args:
            access_token (str): The access token for authentication.
            config_path (str): The path to the configuration file.

        Raises:
            FileNotFoundError: If the specified configuration file does not exist.
        """
        if auth_token is None:
            auth_token = os.environ.get("AUTH_TOKEN")
        if auth_token is None:
            raise KeyError(
                "The auth_token client option must be set either by passing auth token to the client or by setting the AUTH_TOKEN environment variable"
            )
        self.auth_token = auth_token
        try:
            self.config = ConfigReader(config_path=config_path)
        except FileNotFoundError:
            raise FileNotFoundError(f"Config file not found: '{config_path}'")
        self.request_handler = APIRequestHandler(
            auth_token=self.auth_token, config=self.config
        )
        self.headers = {
            "Content-type": "application/json",
            "Accept": "text/plain",
            "access-token": self.auth_token,
        }

    def register(self, register_object: Registrable) -> dict:
        """
        Register an object with the remote API.

        Args:
            register_object: The object to register.

        Returns:
            dict: The response from the API.

        Raises:
            ValueError: If the given object type cannot be registered.
            requests.exceptions.RequestException: If there's an issue with the API request.

        This function registers an object with the remote API based on its type.
        The function checks the type of the register_object and sends the appropriate API request to register it.

        Example usage:
        >>> from gru.sources.datasource import DataSource
        >>> data_source = DataSource(name="example_source", ... )  # Create a DataSource object
        >>> yugen_client.register(data_source)  # Register the DataSource
        """

        if isinstance(register_object, Registrable):
            try:

                url_path, json_req = self.request_handler.create_register_json(
                    register_object=register_object
                )

                response = self.request_handler.send_request(
                    method="post", url_path=url_path, json_data=json_req
                )

                response_handler = APIResponseHandler(response)
                response_handler.check_for_errors()

                data = response_handler.get_success_data()
                print(data)

            except requests.exceptions.RequestException as e:
                print(f"Connection error: {e}")
            except ApiError as e:
                print(f"API error: {e.title} - {e.message}")
        else:
            raise TypeError(
                f"The provided object of type `{register_object.entity_type}` is not Registrable"
            )

    def deploy(
        self,
        name: str,
        entity_type: EntityType,
        workflow: Workflow = Workflow.AIRFLOW,
        dry_run: Optional[bool] = False,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
    ) -> dict:
        """
        Deploy an object with the remote API.

        Args:
            entity (Any): The deployable entity.
            deployment_strategy (DeploymentStrategy, optional): The deployment strategy to use. Defaults to None.
            name (Optional[str], optional): Name for the deployment. Defaults to None.
            dry_run (Optional[bool], optional): Whether to perform a dry run. Defaults to False.
            start_date (Optional[datetime], optional): Deployment start date. Defaults to None.
            end_date (Optional[datetime], optional): Deployment end date. Defaults to None.

        Returns:
            dict: The response from the API.
        """

        deployment_request_type = get_deployment_request_class(workflow)
        deployment_strategy_instance = deployment_request_type(
            workflow_name=name,
            entity_type=entity_type,
            config=self.config,
            dry_run=dry_run,
            start_date=start_date,
            end_date=end_date,
        )

        if deployment_strategy_instance.is_entity_deployable():
            deployment_options = (
                deployment_strategy_instance.create_deployment_options()
            )
            url_path = deployment_options.get("url_path")
            json_req = deployment_options.get("json")

            try:
                response = self.request_handler.send_request(
                    method="post", url_path=url_path, json_data=json_req
                )

                response_handler = APIResponseHandler(response)
                response_handler.check_for_errors()

                data = response_handler.get_success_data()
                print(data)

            except requests.exceptions.RequestException as e:
                print(f"Connection error: {e}")
            except ApiError as e:
                print(f"API error: {e.title} - {e.message}")
        else:
            raise TypeError(
                f"The provided entity type `{entity_type}` is not Depoyable with {workflow.value}"
            )

    def update(
        self,
        update_obj: Union[Type[UpdateFeature], Type[UpdatePreprocessingTransform]],
        entity_type: EntityType,
    ):
        """
        Update an existing object of a specific entity type through the remote API.

        Args:
            update_obj (Union[Type[UpdateFeature], Type[UpdatePreprocessingTransform]]):
                The object representing the updates to be made.
            entity_type (EntityType):
                The type of entity to update. It should be one of the supported EntityType values.

        Returns:
            dict:
                The response from the API containing the results of the update.

        This method allows updating an existing object of a specific entity type through the remote API.
        It checks if the provided entity type is valid for updates and sends the appropriate API request
        to update the object.

        Example usage:
        >>> from gru.features.update_feature import UpdateFeature
        >>> update_feature = UpdateFeature(name="updated_feature", ...)  # Create an UpdateFeature object
        >>> yugen_client.update(update_feature, EntityType.RAW_FEATURE)  # Update the Raw Feature object.
        """
        for type in self.config.update_types:
            if entity_type.value == type:
                try:
                    url_path_id = str(
                        self.config.update_paths[type].replace("{id}", update_obj.name)
                    )
                    print(url_path_id)
                    api_url = urljoin(self.config.stuart_api_endpoint, url_path_id)

                    json_req = update_obj.to_json()

                    json_req["type"] = type
                    json_req[
                        "airflow_dags_backup_path"
                    ] = self.config.airflow_dags_backup_path
                    response = requests.patch(
                        url=api_url,
                        json=json_req,
                        headers=self.headers,
                    )

                    response.raise_for_status()
                    print(response.json())
                    return response.json()
                except requests.exceptions.RequestException as e:
                    raise e
                except Exception:
                    raise ValueError(f"Failed to update the object. ")

        raise ValueError(f"The given object type cannot be updated. ")

    def generate_token(self, generate_token_object: GenerateAccessToken):
        """
        Perform create user and generate access token for users.
        Args:
            generate_token_obj: The object to generate token.

        Returns:
            dict: The response from the API.
        """
        try:
            api_url = urljoin(
                self.config.stuart_api_endpoint,
                self.config.generate_token,
            )
            json_req = generate_token_object.to_json()

            response = requests.post(url=api_url, json=json_req, headers=self.headers)

            response.raise_for_status()
            print(response.json())
            return response.json()
        except requests.exceptions.RequestException as e:
            raise e
        except Exception:
            raise ValueError(
                f"Failed to register user and create access token for: '{generate_token_object.user_email}'."
            )

    def get_feature_health(
        self,
        entity_type: EntityType = EntityType.RAW_FEATURE,
        feature_name: Optional[str] = None,
    ) -> dict:
        """
        Retrieve the health status of a feature.

        Args:
            feature_name (str, optional): The name of the feature. If provided, retrieves the health
                status of the specific feature. If not provided, retrieves the overall health status
                of all features.

        Returns:
            dict: The response from the API.

        Raises:
            requests.exceptions.RequestException: If there's an issue with the API request.
            ValueError: If the feature name is provided but invalid.

        Example usage:
        >>> yugen_client.get_feature_health()  # Retrieve overall health status of all features
        >>> yugen_client.get_feature_health("feature-name1")  # Retrieve health status of a specific feature
        """
        try:
            if entity_type.value in self.config.health_types:
                url_path = self.config.health_paths[entity_type.value]
                if feature_name:
                    url_path = f"{url_path}/{feature_name}"
                response = self.request_handler.send_request(
                    method="GET", url_path=url_path
                )

                if response.status_code == 200:
                    logger.info(f"response: {response.json()}")
                    return response.json()
                else:
                    logger.warning(
                        f"Failed to retrieve feature health: {response.json()}"
                    )
                    return {}
            else:
                raise ValueError("Health data not supported for the given entity type")
        except Exception as e:
            logger.error(f"Failed to retrive feature health: {str(e)}")
            return {}

    def deploy_ai_agent(self, ai_agent: AIAgent) -> Dict:
        """
        Deploy an AI agent with the remote API.

        Args:
            ai_agent (AIAgent): The AIAgent object to deploy.

        Returns:
            dict: The response from the API.

        Raises:
            requests.exceptions.RequestException: If there's an issue with the API request.
            ValueError: If the deployment fails.
        """
        try:
            url_path = self.config.api_endpoints["ai_agent"]["deploy"]
            json_req = ai_agent.to_json()

            response = self.request_handler.send_request(
                method="post", url_path=url_path, json_data=json_req
            )

            response_handler = APIResponseHandler(response)
            response_handler.check_for_errors()

            return response_handler.get_success_data()

        except requests.exceptions.RequestException as e:
            raise ValueError(f"Network error during AI agent deployment: {str(e)}")
        except ApiError as e:
            raise ValueError(f"Failed to deploy AI agent: {e.message}")

    def send_rabbitmq_data(self, ai_agent_name: str, message: AIAgentMessage) -> Dict:
        """
        Send a message to an AI agent.

        Args:
            ai_agent_name (str): The name of the AI agent to send the message to.
            message (AIAgentMessage): The message to send to the AI agent.

        Returns:
            dict: The response from the API.

        Raises:
            requests.exceptions.RequestException: If there's an issue with the API request.
            ValueError: If sending the message fails.
        """
        try:
            url_path = self.config.api_endpoints["ai_agent"]["send_message"]
            json_req = {"ai_agent_name": ai_agent_name, "message": message.to_json()}

            response = self.request_handler.send_request(
                method="post", url_path=url_path, json_data=json_req
            )

            response_handler = APIResponseHandler(response)
            response_handler.check_for_errors()

            return response_handler.get_success_data()

        except requests.exceptions.RequestException as e:
            raise ValueError(
                f"Network error during send message to rabbit MQ: {str(e)}"
            )
        except ApiError as e:
            raise ValueError(f"Failed to send message to rabbit MQ: {e.message}")
    
    def setup_component(self, cluster_name:str, config_file:str):
        try:
            return component_setup(cluster_name, config_file, self.auth_token)
        except FileNotFoundError:
            raise ValueError(f"Error: {config_file} file not found.")
        except ApiError as api_error:
            raise ValueError(api_error.message)
        
    def register_agent(self, agent_folder, cluster_name, image, image_pull_secret):
        try:
            return register_agent(self.auth_token, agent_folder, cluster_name, image, image_pull_secret)
        except FileNotFoundError:
            raise ValueError(f"Error: {os.path.join(agent_folder, AGENT_CONFIG_FILE_NAME)} file not found.")
        except ApiError as api_error:
            raise ValueError(api_error.message)
    
    def deploy_agent(self, agent_name):
        try:
           return deploy_agent(self.auth_token, agent_name)
        except ApiError as api_error:
            raise ValueError(api_error.message)
    
    def prompt_agent(self, agent_name: str, prompt_file: str):
        try:
            prompt_data = read_prompt_file(prompt_file)
            return prompt_agent(agent_name, prompt_data, self.auth_token)
        except FileNotFoundError as e:
            raise ValueError(f"File error: {str(e)}")
        except json.JSONDecodeError as e:
            raise ValueError(f"JSON error: {str(e)}")



from typing import Any, Dict, List, Optional
import os, json, uuid, traceback
from datetime import datetime
from pymilvus import (
    Collection,
    CollectionSchema,
    DataType,
    FieldSchema,
    connections,
    utility
)
from gru.agents.tools.core.vector_db.base import (
    AddToCollectionResponse, 
    CreateCollectionResponse, 
    DeleteFromCollectionResponse,
    DropCollectionResponse,
    GetCollectionInfoResponse,
    ListCollectionsResponse, 
    SimilaritySearchResponse, 
    UpdateCollectionResponse, 
    VectorDBClient, 
    VectorDBConfig
)
from pymilvus.exceptions import DataNotMatchException


class MilvusClient(VectorDBClient):
    MILVUS_TYPES = {
        "INT64": DataType.INT64,
        "VARCHAR": DataType.VARCHAR,
        "FLOAT_VECTOR": DataType.FLOAT_VECTOR,
        "FLOAT": DataType.FLOAT,
        "BOOL": DataType.BOOL
    }

    DEFAULT_INDEX_PARAMS = {
        DataType.FLOAT_VECTOR: {
            "index_type": "IVF_FLAT",
            "params": {
                "nlist": 1024,
                "metric_type": "L2"
            }
        },
        DataType.VARCHAR: {
            "index_type": "Trie",
            "params": {}
        },
        DataType.INT64: {
            "index_type": "STL_SORT",
            "params": {}
        },
        DataType.FLOAT: {
            "index_type": "STL_SORT",
            "params": {}
        }
    }

    def __init__(self):
        self.config = VectorDBConfig(
            host=os.getenv("MILVUS_HOST", "localhost"),
            port=int(os.getenv("MILVUS_PORT", "19530")),
            user=os.getenv("MILVUS_USER", "root"),
            password=os.getenv("MILVUS_PASSWORD", "Milvus"),
            connection_timeout=int(os.getenv("MILVUS_CONNECTION_TIMEOUT", "30")),
            max_retries=int(os.getenv("MILVUS_MAX_RETRIES", "3"))
        )
        self.connect(self.config)

    def connect(self, config:VectorDBConfig) -> None:
        """Initialize connection to Milvus"""
        try:
            connections.connect(
                alias="default",
                host=config.host,
                port=config.port,
                user=config.user,
                password=config.password,
                timeout=config.connection_timeout
            )
        except Exception as e:
            raise ConnectionError(f"Failed to connect to Milvus: {str(e)}")

    def _create_field_schema(self, field_config: Dict[str, Any]) -> FieldSchema:
        return FieldSchema(
            name=field_config["name"],
            dtype=self.MILVUS_TYPES[field_config["dtype"]],
            is_primary=field_config.get("is_primary", False),
            auto_id=field_config.get("auto_id", False),
            max_length=field_config.get("max_length"),
            dim=field_config.get("dim")
        )

    def create_collection_sync(
            self,
            collection_name: str,
            schema: Dict[str, Any],
            replace_existing: bool = False
    ) -> CreateCollectionResponse:
        """Create a new collection with specified schema synchronously"""
        try:
            if utility.has_collection(collection_name):
                if replace_existing:
                    print(f"Dropping existing collection: {collection_name}")
                    utility.drop_collection(collection_name)
                else:
                    return CreateCollectionResponse(
                        status="error",
                        message=f"Collection {collection_name} already exists",
                        collection_name=collection_name
                    )

            field_schemas = []
            for field in schema["fields"]:
                if field["dtype"] == "FLOAT_VECTOR":
                    if "dim" not in field:
                        return CreateCollectionResponse(
                            status="error",
                            message=f"Dimension required for FLOAT_VECTOR field {field['name']}",
                            collection_name=collection_name
                        )
                    field_schema = FieldSchema(
                        name=field["name"],
                        dtype=self.MILVUS_TYPES[field["dtype"]],
                        dim=field["dim"],
                        is_primary=field.get("is_primary", False),
                        auto_id=field.get("auto_id", False)
                    )
                else:
                    field_schema = FieldSchema(
                        name=field["name"],
                        dtype=self.MILVUS_TYPES[field["dtype"]],
                        is_primary=field.get("is_primary", False),
                        auto_id=field.get("auto_id", False),
                        max_length=field.get("max_length")
                    )
                field_schemas.append(field_schema)

            collection_schema = CollectionSchema(
                fields=field_schemas,
                description=schema.get("description", "")
            )

            collection = Collection(
                name=collection_name,
                schema=collection_schema
            )

            explicit_indexes = schema.get("create_indexes", [])
            if explicit_indexes:
                print(f"Creating explicitly defined indexes for collection {collection_name}")
                
                for index_def in explicit_indexes:
                    field_name = index_def.get("field_name")
                    index_params = index_def.get("index_params", {})
                    
                    if not field_name:
                        print("Skipping index with no field_name specified")
                        continue
                        
                    print(f"Creating index for field {field_name}")
                    collection.create_index(
                        field_name=field_name,
                        index_params=index_params
                    )
            collection.load()
            return CreateCollectionResponse(
                status="success",
                message=f"Collection {collection_name} created successfully",
                collection_name=collection_name
            )
        except Exception as e:
            return CreateCollectionResponse(
                status="error",
                message=f"Failed to create collection: {str(e)}",
                collection_name=collection_name
            )

    async def create_collection(
        self,
        collection_name: str,
        schema: Dict[str, Any],
        replace_existing: bool = False
    ) -> CreateCollectionResponse:
        try:
            if utility.has_collection(collection_name):
                if replace_existing:
                    utility.drop_collection(collection_name)
                else:
                    return CreateCollectionResponse(
                        status="error",
                        message=f"Collection {collection_name} already exists",
                        collection_name=collection_name
                    )

            field_schemas = [
                self._create_field_schema(field) 
                for field in schema["fields"]
            ]
            collection_schema = CollectionSchema(
                fields=field_schemas,
                description=schema.get("description", "")
            )
            
            collection = Collection(
                name=collection_name, 
                schema=collection_schema
            )

            # Create indexes for each field based on its type
            for field_schema in field_schemas:
                index_params = self.DEFAULT_INDEX_PARAMS.get(field_schema.dtype)
                if index_params:
                    collection.create_index(
                        field_name=field_schema.name,
                        index_params=index_params
                    )
            collection.load()
            return CreateCollectionResponse(
                status="success",
                message=f"Collection {collection_name} created successfully",
                collection_name=collection_name
            )
        except Exception as e:
            return CreateCollectionResponse(
                status="error",
                message=f"Failed to create collection: {str(e)}",
                collection_name=collection_name
            )

    async def add_to_collection(
        self, 
        collection_name: str,
        embeddings: List[List[float]],
        data: List[Dict[str, Any]]
    ) -> AddToCollectionResponse:
        try:
            collection = Collection(collection_name)
            schema = collection.schema
            field_names = {field.name for field in schema.fields}  

            data_to_insert = []
            for embedding, meta in zip(embeddings, data):
                doc = {
                    "id": meta.get("memory_id", str(uuid.uuid4())),
                    "embedding": embedding
                }

                for field_name in field_names:
                    if field_name == "memory_id" or field_name == "embedding":
                        continue  

                    if field_name in meta:
                        value = meta.get(field_name)
                        doc[field_name] = json.dumps(value) if isinstance(value, dict) else value

                if "created_at" not in doc:
                    doc["created_at"] = datetime.utcnow().isoformat()

                data_to_insert.append(doc)

            print(f"Inserting {len(data_to_insert)} documents into {collection_name}")
            result = collection.insert(data_to_insert)

            return AddToCollectionResponse(
                status="success",
                message=f"Inserted {len(data_to_insert)} documents into {collection_name}",
                ids=result.primary_keys
            )
        except DataNotMatchException as e:
            return AddToCollectionResponse(
                status="error",
                message=f"Data does not match schema: {str(e)}",
                ids=[]
            )
        except Exception as e:
            print(f"Error in add_to_collection: {str(e)}")
            traceback.print_exc()
            return AddToCollectionResponse(
                status="error",
                message=f"Failed to insert data: {str(e)}",
                ids=[]
            )

    async def update_collection(
            self,
            collection_name: str,
            memory_id: str,
            embedding: List[float],
            data: Optional[Dict[str, Any]] = None
    ) -> UpdateCollectionResponse:
        try:
            collection = Collection(collection_name)

            # Get the schema to ensure consistent field handling
            schema = collection.schema
            field_names = {field.name for field in schema.fields}

            # First verify the document exists
            search_expr = f'id == "{memory_id}"'
            result = collection.query(
                expr=search_expr,
                output_fields=["id"]
            )

            if not result:
                return UpdateCollectionResponse(
                    status="error",
                    message=f"Document with id {memory_id} not found",
                    id=memory_id,
                    updated_count=0
                )

            # Delete existing document
            collection.delete(f"id == '{memory_id}'")

            # Construct new document with schema-aware fields
            doc = {
                "id": memory_id,
                "embedding": embedding,
                "text": data.get("text", ""),
                "created_at": data.get("created_at", datetime.now().isoformat())
            }


            meta_dict = data.get("data", {})

            for field in schema.fields:
                field_name = field.name
                if field_name not in doc and field_name in meta_dict:
                    doc[field_name] = meta_dict[field_name]

            # Insert updated document
            insert_result = collection.insert([doc])

            return UpdateCollectionResponse(
                status="success",
                message="Document updated successfully",
                id=memory_id,
                updated_count=insert_result.insert_count
            )

        except Exception as e:
            return UpdateCollectionResponse(
                status="error",
                message=f"Failed to update document: {str(e)}",
                id=memory_id,
                updated_count=0
            )

    async def delete_from_collection(
        self,
        collection_name: str,
        memory_id: str
    ) -> DeleteFromCollectionResponse:
        try:
            collection = Collection(collection_name)
            delete_expr = f"id == '{memory_id}'"
            result = collection.delete(delete_expr)
            
            return DeleteFromCollectionResponse(
                status="success",
                message=f"Deleted {result.delete_count} documents from {collection_name}",
            )
        except Exception as e:
            return DeleteFromCollectionResponse(
                status="error",
                message=f"Failed to delete document: {str(e)}"
            )
        
    async def similarity_search(
        self,
        collection_name: str,
        query_vector: List[float],
        anns_field: Optional[str] = None,
        top_k: int = 5,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> SimilaritySearchResponse:
        """
        Performs a similarity search using the query vector
        and returns top-k results.
        """
        collection = Collection(collection_name)

        # If anns_field is not provided, use the first float vector field
        if anns_field is None:
            schema = collection.schema
            for field in schema.fields:
                if field.dtype == DataType.FLOAT_VECTOR:
                    anns_field = field.name
                break  
        
        # If output_fields is not provided, use all fields except anns_field
        if output_fields is None:
            schema = collection.schema
            output_fields = [field.name for field in schema.fields 
                            if field.dtype != DataType.FLOAT_VECTOR]

        if search_params is None:
            search_params = {
                "metric_type":  "L2",
                "params": {
                    "nprobe": 10
                },
            }

        # Perform the search
        try:
            results = collection.search(
            data=[query_vector], 
            anns_field=anns_field,
            param=search_params,
            limit=top_k,
            output_fields=output_fields,
        )
        except Exception as e:
            print(f"Error in similarity_search: {str(e)}")
            return SimilaritySearchResponse(
                status="error",
                message=f"Failed to search: {str(e)}",
                results=[]
            )

        # Process and format the results
        search_results = []
        for hits in results:
            for hit in hits:
                result = {"score": hit.score}
                if output_fields:
                    for field in output_fields:
                        result[field] = hit.entity.get(field)
                search_results.append(result)

        return SimilaritySearchResponse(
            status="success",
            message="Search completed successfully",
            results=search_results
        )

    async def filtered_search(
        self,
        collection_name: str,
        filter_expr: str,
        limit: int = 5,
        output_fields: Optional[List[str]] = None,
    ) -> SimilaritySearchResponse:
        """Performs a filtered search based on additional query parameters."""
        collection = Collection(collection_name)

        results = collection.query(
            expr=filter_expr, output_fields=output_fields, limit=limit
        )

        return SimilaritySearchResponse(
            status="success",
            message="Search completed successfully",
            results=results
        )

    async def list_collections(self) -> ListCollectionsResponse:
        try:
            print("Inside milvus client")
            return ListCollectionsResponse(
                status="success",
                message="Collections listed successfully",
                collections=utility.list_collections()
            )
        except Exception as e:
            print(f"Failed to list collections: {str(e)}")
            return ListCollectionsResponse(
                status="error",
                message=f"Failed to list collections: {str(e)}",
                collections=[]
            )
    
    # TODO reuse in Canso_memory
    async def get_collection_info(
        self,
        collection_name: str
    ) -> GetCollectionInfoResponse:
        """Get collection metadata and statistics"""
        try:
            if not utility.has_collection(collection_name):
                return GetCollectionInfoResponse(
                    status="error",
                    message=f"Collection {collection_name} does not exist",
                    collection_info={}
                )
                
            collection = Collection(collection_name)
            
            fields_info = []
            for field in collection.schema.fields:
                field_info = {
                    "name": field.name,
                    "dtype": str(field.dtype),
                    "is_primary": field.is_primary
                }
                if hasattr(field, "dim") and field.dim is not None:
                    field_info["dim"] = field.dim
                if hasattr(field, "max_length") and field.max_length is not None:
                    field_info["max_length"] = field.max_length
                    
                fields_info.append(field_info)
            
            index_info = []
            for index in collection.indexes:
                index_info.append({
                    "field_name": index.field_name,
                    "index_type": index.params.get("index_type", ""),
                    "params": index.params.get("params", {})
                })
            
            return GetCollectionInfoResponse(
                status="success",
                message="Collection info retrieved successfully",
                collection_info={
                    "name": collection_name,
                    "entity_count": collection.num_entities,
                    "fields": fields_info,
                    "indexes": index_info,
                    "description": collection.schema.description
                }
            )
        except Exception as e:
            print(f"Error in get_collection_info: {str(e)}")
            traceback.print_exc()
            return GetCollectionInfoResponse(
                status="error",
                message=f"Failed to get collection info: {str(e)}",
                collection_info={}
            )

    async def drop_collection(
        self,
        collection_name: str
    ) -> DropCollectionResponse:
        try:
            utility.drop_collection(collection_name)
            return DropCollectionResponse(
                status="success",
                message=f"Collection {collection_name} dropped successfully"
            )
        except Exception as e:
            return DropCollectionResponse(
                status="error",
                message=f"Failed to drop collection: {str(e)}"
            )



from gru.agents.tools.core.llm_client.base import LLMClient
from openai import OpenAI
from typing import Optional
import os


class OpenAILLMClient(LLMClient):
    def __init__(
        self,
        api_key: Optional[str] = None,
        model: str = "gpt-4",
        temperature: float = 0,
    ):
        self.client = OpenAI(api_key=os.getenv("OPENAI_API_KEY"))
        self.model = model
        self.temperature = temperature

    async def generate(
        self,
        system_prompt: str,
        user_prompt: str
    ) -> Optional[str]:
        try:
            response = self.client.chat.completions.create(
                model=self.model,
                temperature=self.temperature,
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_prompt},
                ],
            )
            return response.choices[0].message.content
        except Exception as e:
            raise Exception(f"Error generating LLM response: {str(e)}")



from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional, Tuple
from gru.agents.tools.core.vector_db.base import VectorDBClient
from gru.agents.tools.core.code_generator.models import RetrievalResult


class ContextRetriever(ABC):
    def __init__(
        self,
        vector_store: VectorDBClient,
        system_prompt: Optional[str] = None,
        user_prompt_template: Optional[str] = None,
    ):
        """
        Base class for Context Retrievers.
        """
        self.vector_store = vector_store
        self.system_prompt = system_prompt
        self.user_prompt_template = user_prompt_template

    @abstractmethod
    async def get_relevant_datasources_and_schemas(
        self,
        query_embedding: List[float],
        top_k: int,
        embedding_field: str,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> Tuple[List[str], List[str]]:
        """
        Retrieves relevant data sources and their schemas.
        """
        pass

    @abstractmethod
    async def get_similar_examples(
        self,
        query_embedding: List[float],
        top_k: int,
        embedding_field: str,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> List[str]:
        """Retrieves similar examples based on the query embedding."""
        pass

    @abstractmethod
    async def get_documentation(
        self,
        query_embedding: List[float],
        top_k: int,
        embedding_field: str,
        search_params: Optional[Dict[str, Any]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> List[str]:
        """Retrieves relevant documentation based on the query embedding."""
        pass

    @abstractmethod
    async def retrieve_context(
        self, query_embedding: List[float], top_k: int
    ) -> RetrievalResult:
        """Retrieves all relevant context for the query embedding."""
        pass



import yaml
from pathlib import Path
from typing import Optional
from gru.models.common.hpa_config import HPAConfig
from gru import DEFAULT_CANSO_API_CONFIG


class ConfigReader:
    """
    Reads and provides access to configuration data from a YAML file.
    Handles both platform configuration and HPA configuration.

    Attributes:
        config_path (str): The path to the YAML configuration file.
        stuart_api_endpoint (str): The base API endpoint.
        register_class_map (dict): Dictionary mapping classes to their registration API paths.
        deploy_class_map (dict): Dictionary mapping classes to their deployment API paths.
        register_classes (list): List of classes that can be registered.
        deploy_classes (list): List of classes that can be deployed.
    """

    def __init__(self, config_path: str):
        """
        Initialize the ConfigReader.

        Args:
            config_path (str): The path to the YAML configuration file
        """
        self.config_path = config_path
        self.read_config()

    def read_config(self):
        """
        Read and parse the configuration file based on config type.
        """
        try:
            config_dict = self._read_file(self.config_path)
            if "api_endpoints" in config_dict:
                self.set_platform_endpoints(config_dict["api_endpoints"])
            if "processing_engine_configs" in config_dict:
                self.validate_processing_engine_configs(config_dict)
                self.set_processing_engine_configs(
                    config_dict["processing_engine_configs"]
                )

            for attr_name, attr_value in config_dict.items():
                setattr(self, attr_name, attr_value)
        except KeyError as e:
            raise KeyError(f"Key not found in configuration file: '{e.args[0]}'")

    @classmethod
    def _read_file(self, file_path: str):
        try:
            with open(file_path, "r") as config_file:
                config_dict = yaml.safe_load(config_file)
            return config_dict
        except FileNotFoundError:
            raise FileNotFoundError(f"Config file not found: '{file_path}'")

    def set_platform_endpoints(self, platform_endpoints):
        """
        Set the registration and deployment API paths based on the configuration.

        Args:
            api_paths (dict): Dictionary containing registration and deployment API paths.

        Raises:
            KeyError: If the required keys are not found in the API paths.
        """
        try:
            self.stuart_api_endpoint = platform_endpoints["stuart"]
            self.risk_mgmt_endpoint = platform_endpoints["risk_mgmt"]
            self.agents_service_url = platform_endpoints["agents"]
            self.online_fs_service_url = platform_endpoints["online_fs_service"]

            self.register_paths = platform_endpoints["register"]
            self.deploy_paths = platform_endpoints["deploy"]
            self.dry_run_paths = platform_endpoints["dry_run"]
            self.update_paths = platform_endpoints["update"]
            self.health_paths = platform_endpoints["health"]
            self.risk_paths = platform_endpoints.get("risk", None)

            self.register_types = list(self.register_paths.keys())
            self.deploy_types = list(self.deploy_paths.keys())
            self.dry_run_types = list(self.dry_run_paths.keys())
            self.update_types = list(self.update_paths.keys())
            self.health_types = list(self.health_paths.keys())

            self.generate_token = platform_endpoints["generate_token"]

            self.component_setup_endpoint = platform_endpoints["component"]["setup"]
            self.agent_register_endpoint = platform_endpoints["ai_agent"]["register"]
            self.agent_deploy_endpoint = platform_endpoints["ai_agent"]["deploy"]
            self.agent_update_endpoint = platform_endpoints["ai_agent"]["update"]
            self.agent_delete_endpoint = platform_endpoints["ai_agent"]["delete"]

            ai_agent_endpoints = platform_endpoints.get("ai_agent", {})
            self.agent_prompt_endpoint = ai_agent_endpoints.get("prompt")
            self.agent_converse_endpoint = ai_agent_endpoints.get("converse")
            self.agent_memory_endpoint = ai_agent_endpoints.get("memory")

            if self.risk_paths:
                self.workflow_list_endpoint = self.risk_paths["list"]
                self.workflow_get_endpoint = self.risk_paths["get"]
                self.workflow_create_endpoint = self.risk_paths["create"]
                self.workflow_update_status_endpoint = self.risk_paths["update_status"]
                self.workflow_rules_list_endpoint = self.risk_paths["rules"]["list"]
                self.workflow_rules_create_endpoint = self.risk_paths["rules"]["create"]
                self.workflow_deploy_endpoint = self.risk_paths["deploy"]

        except KeyError as e:
            raise KeyError(f"Key not found in the API paths: '{e.args[0]}'")

    def validate_processing_engine_configs(self, config_dict):
        valid_config_keys = {
            "deploy-mode",
            "spark.kubernetes.driver.memory",
            "spark.kubernetes.executor.memory",
            "spark.kubernetes.authenticate.driver.serviceAccountName",
            "spark.hadoop.fs.s3a.aws.credentials.provider",
            "spark.kubernetes.container.image.pullSecrets",
            "spark.kubernetes.file.upload.path",
            "spark.kubernetes.namespace",
            "num-executors",
            "checkpoint_path",
            "spark.executor.cores",
            "spark.driver.cores",
            "spark.kubernetes.driver.limit.cores",
            "spark.kubernetes.executor.limit.cores",
            "spark.executor.instances",
            "spark.kubernetes.container.image",
        }
        invalid_configs = (
            set(config_dict["processing_engine_configs"].keys()) - valid_config_keys
        )
        if invalid_configs:
            raise ValueError(
                f"Invalid keys in the provided config file: {', '.join(invalid_configs)}. "
                f"Options supported for spark-submit on K8S are: {', '.join(valid_config_keys)}"
            )

    def set_processing_engine_configs(self, processing_engine_configs):
        try:
            self.deploy_mode = processing_engine_configs["deploy-mode"]
            self.driver_memory = processing_engine_configs[
                "spark.kubernetes.driver.memory"
            ]
            self.executor_memory = processing_engine_configs[
                "spark.kubernetes.executor.memory"
            ]
            self.service_account = processing_engine_configs[
                "spark.kubernetes.authenticate.driver.serviceAccountName"
            ]
            self.aws_credential_provider = processing_engine_configs[
                "spark.hadoop.fs.s3a.aws.credentials.provider"
            ]
            self.container_image_pull_secrets = processing_engine_configs[
                "spark.kubernetes.container.image.pullSecrets"
            ]
            self.container_image = processing_engine_configs.get(
                "spark.kubernetes.container.image"
            )
            self.file_upload_path = processing_engine_configs[
                "spark.kubernetes.file.upload.path"
            ]
            self.name_space = processing_engine_configs["spark.kubernetes.namespace"]
            self.num_executors = processing_engine_configs["num-executors"]
            self.checkpoint_path = processing_engine_configs["checkpoint_path"]
            self.executor_cores = processing_engine_configs["spark.executor.cores"]
            self.driver_cores = processing_engine_configs["spark.driver.cores"]
            self.driver_limit_cores = processing_engine_configs[
                "spark.kubernetes.driver.limit.cores"
            ]
            self.executor_limit_cores = processing_engine_configs[
                "spark.kubernetes.executor.limit.cores"
            ]
            self.executor_instances = processing_engine_configs[
                "spark.executor.instances"
            ]
        except KeyError as e:
            raise KeyError(
                f"Key not found in the Proessing Engine Configs: '{e.args[0]}'"
            )

    @classmethod
    def load_hpa_config(self, file_path: Optional[str] = None) -> Optional[HPAConfig]:
        """
        Load HPA configuration from file or default

        Args:
            file_path: Optional path to custom HPA config file

        Returns:
            HPAConfig object or None if no config provided

        Raises:
            ValueError: If config file is invalid or cannot be read
        """
        try:
            if file_path is None:
                file_path = str(
                    Path(DEFAULT_CANSO_API_CONFIG).parent
                    / "config"
                    / "default_hpa_configs.yaml"
                )

            config_dict = self._read_file(file_path)
            hpa_config = HPAConfig(**config_dict)
            return hpa_config
        except Exception as e:
            raise ValueError(f"Failed to load HPA config: {str(e)}")



"""
Sink Module for Object Storage
"""

from typing import Dict, Optional, Any, Union
from gru.sinks.sink import Sink
from gru.sinks.sink_family import SinkFamily
from gru.utils.cloud_providers import CloudProviders
from gru.sources.file_types import CSVType, ParquetType
from datetime import datetime


class ObjectStorageSink(Sink):
    """A Sink which writes data to Object Storage"""

    def __init__(
        self,
        name: str,
        uri_scheme: str,
        uri_scheme_spark: str,
        bucket: str,
        leading_key: str,
        file_type: Union[CSVType, ParquetType],
        ts_suffix_key: str = "%Y-%m-%d-%H-%M",
        file_type_properties: Optional[Dict[str, Any]] = {},
        created_at=datetime.now(),
        **kwargs,
    ) -> None:
        """
        Initialize the ObjectStorageSink

        :param bucket: The bucket to write files to.
        :param leading_key: The key to write files to.
        :param file_type: The format of the file i.e. CSV, Parquet etc
        :param file_type_properties: The properties of the file format.
        :param ts_suffix_key: An optional timestamped suffix
            (e.g. %Y-%m-%d-%H-%M or d=%Y-%m-%d/h=%H-%M)
        """
        self._check_non_empty_bucket_and_key(bucket, leading_key)

        super().__init__(
            name=name,
            family=SinkFamily.OBJECT_STORAGE.value,
            physical_uri={
                "bucket": bucket,
                "leading_key": leading_key,
                "ts_suffix_key": ts_suffix_key,
            },
            **kwargs,
        )
        self.uri_scheme = uri_scheme
        self.uri_scheme_spark = uri_scheme_spark
        self.file_type = file_type
        self.file_type_properties = file_type_properties

        self.physical_uri["uri_scheme"] = self.uri_scheme
        self.physical_uri["uri_scheme_spark"] = self.uri_scheme_spark
        self.created_at = created_at

    @classmethod
    def _check_non_empty_bucket_and_key(cls, bucket: str, leading_key: str) -> None:
        """
        Validate that the bucket and leading key are not empty.

        :param bucket: The bucket name.
        :param leading_key: The leading key.
        :raises ValueError: If the bucket or leading key is empty.
        """
        if not bucket:
            raise ValueError("Bucket cannot be empty")
        if not leading_key:
            raise ValueError("Leading key cannot be empty")

    def to_json(self) -> Dict:
        """
        Convert the ObjectStorageSink object to a JSON dictionary.

        :return: The JSON representation of the object.
        """
        return {
            "name": self.name,
            "family": self.family,
            "physical_uri": self.physical_uri,
            "type": self.type,
            "file_type": self.file_type.type if self.file_type else None,
            "file_type_properties": self.file_type.options if self.file_type else None,
            "cloud_provider": self.cloud_provider.value,
            "description": self.description,
            "metadata": self.metadata,
            "owners": self.owners,
            "created_at": str(self.created_at),
        }

    @classmethod
    def from_json(cls, json_dict: Dict) -> "ObjectStorageSink":
        """
        Create an ObjectStorageSink object from a JSON dictionary.

        :param json_dict: The JSON representation of the object.
        :return: The ObjectStorageSink object.
        :raises KeyError: If any required key is missing from the JSON dictionary.
        :raises ValueError: If the bucket or leading key is empty.
        """
        uri_scheme = json_dict["uri_scheme"]
        uri_scheme_spark = json_dict["uri_scheme_spark"]
        bucket = json_dict["bucket"]
        leading_key = json_dict["leading_key"]
        ts_suffix_key = json_dict["ts_suffix_key"]
        file_type = json_dict["file_type"]
        file_type_properties = json_dict["file_type_properties"]

        cls._check_non_empty_bucket_and_key(bucket, leading_key)

        return ObjectStorageSink(
            uri_scheme=uri_scheme,
            uri_scheme_spark=uri_scheme_spark,
            bucket=bucket,
            leading_key=leading_key,
            ts_suffix_key=ts_suffix_key,
            file_type=file_type,
            file_type_properties=file_type_properties,
        )


class S3Sink(ObjectStorageSink):
    """
    A Sink which writes data to AWS S3.

    Parameters:
    - bucket (str): The name of the bucket to write files to.
    - leading_key (str): The leading key to write files to.
    - file_type (str): The format of the file, e.g., CSV, Parquet, etc.
    - ts_suffix_key (str): An optional timestamped suffix for file names (default: "%Y-%m-%d-%H-%M").
    - file_type_properties (Optional[Dict[str, Any]]): The properties of the file format (default: None).

    Inherits:
    - ObjectStorageSink: The base class for writing data to object storage.

    Attributes:
    - cloud_provider (str): The cloud provider.
    - uri_scheme (str): The URI scheme for S3.
    - uri_scheme_spark (str): The URI scheme for Spark to access S3.
    """

    def __init__(
        self,
        name: str,
        bucket: str,
        leading_key: str,
        file_type: Union[CSVType, ParquetType],
        ts_suffix_key: str = "%Y-%m-%d-%H-%M",
        cloud_provider=CloudProviders.AWS,
        file_type_properties: Optional[Dict[str, Any]] = None,
        created_at=datetime.now(),
        **kwargs,
    ) -> None:
        super().__init__(
            name=name,
            uri_scheme="s3",
            uri_scheme_spark="s3a",
            bucket=bucket,
            leading_key=leading_key,
            file_type=file_type,
            ts_suffix_key=ts_suffix_key,
            file_type_properties=file_type_properties,
            type="S3Sink",
            **kwargs,
        )
        self.leading_key = leading_key
        self.cloud_provider = cloud_provider
        self.file_type = file_type
        self.file_type_properties = file_type_properties
        self.created_at = created_at



from enum import Enum
from gru.schema.deployment_strategy import AirflowDeploymentRequest

class Workflow(Enum):
    AIRFLOW = "airflow"

DeploymentRequestMapping = {
    Workflow.AIRFLOW: AirflowDeploymentRequest
}

def get_deployment_request_class(workflow : Workflow):
    if  workflow in DeploymentRequestMapping:
        return DeploymentRequestMapping.get(workflow)
    else:
        raise TypeError(f"workflow type {workflow} not supported yet")




import pytest
from gru.utils.processing_engine import ProcessingEngine


def test_processing_engines():
    """
    Unittest to validate all possible processing engines
    """
    assert ProcessingEngine.PYSPARK_K8S.value == "pyspark_k8s"
    assert ProcessingEngine.PYTHON.value == "python"
    assert ProcessingEngine.DASK.value == "dask"


def test_supported_processing_engines():
    """
    Unittest to validate processing engines currently supported by Yugen.ai platform
    """
    assert ProcessingEngine.get_key("pyspark_k8s") == ProcessingEngine.PYSPARK_K8S


def test_unsupported_processing_engines():
    """
    Unittest to validate the platforms that are not supported by Yugen.ai platform
    """
    with pytest.raises(
        ValueError,
        match="python isn't supported yet. Support expected to be available around Feb, 2024.",
    ):
        ProcessingEngine.get_key("python")

    with pytest.raises(
        ValueError,
        match="dask isn't supported yet. Support expected to be available around Feb, 2024.",
    ):
        ProcessingEngine.get_key("dask")


def test_unexpected_processing_engines():
    """
    Unittest to validate the unexpected/unwanted processing engine
    """
    with pytest.raises(
        ValueError,
        match=f"""Unexpected engine foo. Supported Engines are {', '.join(engine.value for engine in ProcessingEngine)}.""",
    ):
        ProcessingEngine.get_key("foo")



"""
In Machine Leaning, Registry is the store house of all the registered entities. 
Namely:
    - Features
    - Training Data
    - Preprocessing Transforms
    - Data Sources & Data Sinks
"""

import requests

class Registry:
    def __init__(self, access_token):
        self.access_token = access_token

    def get_registered(
            self,
            entity_url: str,
            ) -> list[str]:
        """
        Function to return the list of all entities registered in our platform registry database.
        """

        result = requests.get(
            url=entity_url,
            json={"access_token": self.access_token},
            headers={"Accept": "text/plain", "Content-type": "application/json",},
        )
        return result.json()


    def get_registered_features(self):
        """Function to get the list of all the registered features from the registry database"""
        pass

    def get_registered_data_sources(self):
        """Function to get the list of all the registered data sources from the registry database"""
        pass

    def get_registered_data_sinks(self):
        """Function to get the list of all the registered data sinks from the registry database"""
        pass

    def get_registered_ppt(self):
        """Function to get the list of all the registered Pre Processing Transforms from the registry database"""
        pass

    def get_registered_training_data(self):
        """Function to get the list of all the registered Training Data from the registry database"""
        pass



import base64
import os
import ssl
import threading

import pika
import logging

logger = logging.getLogger(__name__)


class RabbitMQPublisher(threading.Thread):

    CANSO_VHOST = "/canso"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.daemon = True
        self.is_running = True
        self.name = "RabbitMQPublisher"

        self.host = os.environ.get("QUEUE_HOST")
        self.port = int(os.environ.get("QUEUE_PORT", 5671))
        self.username = os.getenv("QUEUE_USERNAME")
        self.password = base64.b64decode(os.getenv("QUEUE_PASSWORD")).decode()

        self.connection = self.get_connection()
        self.channel = self.connection.channel()

    def get_connection(self):
        ssl_context = ssl.create_default_context()
        credentials = pika.PlainCredentials(self.username, self.password)
        parameters = pika.ConnectionParameters(
            host=self.host,
            port=self.port,
            credentials=credentials,
            ssl_options=pika.SSLOptions(ssl_context),
            virtual_host=self.CANSO_VHOST,
        )
        return pika.BlockingConnection(parameters)

    def run(self):
        while self.is_running:
            self.connection.process_data_events(time_limit=1)

    def _publish(
        self, exchange: str, routing_key: str, message: str, correlation_id: str = None
    ):
        self.channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            body=message,
            properties=pika.BasicProperties(
                delivery_mode=2,
                content_type="application/json",
                correlation_id=correlation_id,
            ),
        )

    def publish(
        self, exchange: str, routing_key: str, message: str, correlation_id: str = None
    ):
        self.connection.add_callback_threadsafe(
            lambda: self._publish(exchange, routing_key, message, correlation_id)
        )

    def stop(self):
        self.is_running = False
        if self.connection.is_open:
            # Wait until all the data events have been processed
            self.connection.process_data_events(time_limit=1)
            self.connection.close()



from abc import ABC, abstractmethod
from gru.agents.tools.core.code_generator.models import (
    QueryIntent, RetrievalResult
)


class CodeGenerator(ABC):
    def __init__(self, llm_client):
        self.llm_client = llm_client

    @abstractmethod
    async def generate_code(
        self, query: str, intent: QueryIntent, context: RetrievalResult
    ) -> str:
        pass




class ResponseHandler():
    """
    Class for handling responses.

    Methods:
        - format_response: Abstract method for formatting responses.
    """

    def format_response(self, result):
        """
        Format responses.

        Parameters:
            - result: Result data to be included in the response.

        Returns:
            Any: Formatted response.
        """
        pass


from enum import Enum

class CloudProviders(Enum):
    """
    Enum class representing various cloud providers.
    """

    AWS = "aws"
    GCP = "gcp"
    Azure = "azure"
    IBM = "ibm"
    Oracle = "oci"
    RedisCloud = "redis_cloud"


    @classmethod
    def get_key(cls, value: str):
        """
        Retrieve the key (name) of a cloud provider based on its string value.

        Args:
            value: The string value representing a cloud provider.
        
        Returns:
            str: The key (name) of the managed cloud provider.

        Raises:
            ValueError: If no key is found for the given value.

        Example:
        >>> CloudProviders.get_key("gcp")
        'GCP'
        """
        if value == cls.AWS.value:
            return cls.AWS
        else:
            raise ValueError(f"{value} is in the Platform roadmap but isn't supported yet.")


import unittest

from gru.features.aggregations import SlidingWindowAggregation

obj = SlidingWindowAggregation(
    function="AVG",
    partition_by="provider",
    order_by=["project_id"],
    rangeBetween={"frame_start": 1, "frame_end": 6},
)


def _validate_frame_range_valid_window(
    frame_start_range_exists,
    frame_end_range_exists,
    frame_start_range_val,
    frame_end_range_val,
):
    """
    Validate that the  start range value  should be less than equal to  end range value .

    Raises:
        ValueError: If frame_start_range is greater frame_end_range is None .

    Returns:
        dict

    """
    if (
        frame_start_range_exists
        and frame_end_range_exists
        and (frame_start_range_val <= frame_end_range_val)
    ):
        return {
            "function": "AVG",
            "partition_by": "provider",
            "order_by": ["order_by"],
            "rangeBetween": {"frame_start": 1, "frame_end": 2},
        }
    else:
        raise ValueError(
            "The start value of the range should be smaller than or equal to the end value of the range."
        )


class Test(unittest.TestCase):
    def test_validate_order_by1(self):
        """
        Unittest to ensure order_by list is not empty
        """
        order_by = ["project_id"]
        result = obj._validate_order_by(order_by)
        self.assertEqual(result, True)

    def test_validate_order_by2(self):
        """
        Unittest to check if order_by is empty
        """
        order_by = []
        with self.assertRaises(ValueError):
            result = obj._validate_order_by(order_by)

    def test_validate_range1(self):
        """
        Unittest to validate the range for range_between
        """
        result = obj._validate_range()
        self.assertEqual(result, True)

    def test_validate_frame_range_valid_window2(self):
        """
        Evaluating if  value error is raised given any one of parameter is False.
        """
        with self.assertRaises(ValueError):
            result = _validate_frame_range_valid_window(True, False, 1, 2)

    def test_validate_frame_range_valid_window2(self):
        """
        Checking if the return value is dict or not
        """
        transform_args = {
            "function": "AVG",
            "partition_by": "provider",
            "order_by": ["order_by"],
            "rangeBetween": {"frame_start": 1, "frame_end": 2},
        }
        result = _validate_frame_range_valid_window(True, True, 1, 2)
        self.assertDictEqual(result, transform_args)

    def test_validate_frame_range_valid_window3(self):
        """
        Evaluating if  value error is raised given any one of parameter is False
        """
        with self.assertRaises(ValueError):
            result = _validate_frame_range_valid_window(False, False, 1, 2)


if __name__ == "__main__":
    unittest.main()



from pydantic import BaseModel
from typing import List, Optional


class QueryIntent(BaseModel):
    entities: List[str]
    domains: List[str]


class RetrievalResult(BaseModel):
    tables: List[str]
    schemas: List[str]
    documentation: List[str]
    examples: List[str]
    low_cardinality_values: List[str]
    domain_knowledge: List[str]
    opt_rules: List[str]


class SQLQueryResult(BaseModel):
    sql_query: str
    tables_used: Optional[List[str]] = None
    columns_used: Optional[List[str]] = None
    join_conditions: Optional[List[str]] = None



from typing import Dict, Optional
from gru.utils.config_reader import ConfigReader
import re

class ProcessingEngineConfigs:
    """
    A class representing a feature retrieval service with a Processing Engine Configs.

    Args:
        config_path: A path to config file containing default parameters for processing engine

    Example:
        Initialize ProcessingEngineConfig:
        processing_engine_configs = ProcessingEngineConfigs(config_path="/path/to/config.yaml")
    """

    def __init__(
        self,
        deploy_mode: Optional[str] = "",
        driver_memory: Optional[str] = "",
        executor_memory: Optional[str] = "",
        service_account: Optional[str] = "",
        aws_credential_provider: Optional[str] = "",
        container_image: Optional[str] = "",
        container_image_pull_secrets: Optional[str] = "",
        file_upload_path: Optional[str] = "",
        name_space: Optional[str] = "",
        num_executors: Optional[int] = None,
        checkpoint_path: Optional[str] = None,
        executor_cores: Optional[int] = None,
        driver_cores: Optional[int] = None,
        driver_limit_cores: Optional[str] = "",
        executor_limit_cores: Optional[str] = "",
        executor_instances: Optional[int] = None,
        config_path: Optional[str] = None,
    ):
        self.deploy_mode = deploy_mode
        self.driver_memory = driver_memory
        self.executor_memory = executor_memory
        self.service_account = service_account
        self.aws_credential_provider = aws_credential_provider
        self.container_image = container_image
        self.container_image_pull_secrets = container_image_pull_secrets
        self.file_upload_path = file_upload_path
        self.name_space = name_space
        self.num_executors = num_executors
        self.executor_cores = executor_cores
        self.driver_cores = driver_cores
        self.driver_limit_cores = driver_limit_cores
        self.executor_limit_cores = executor_limit_cores
        self.executor_instances = executor_instances
        self.config = ConfigReader(config_path=config_path or getattr(self, 'default_config_path', None))
        self.checkpoint_path = checkpoint_path or self.config.checkpoint_path
        self._validate_checkpoint_path()

    def _validate_checkpoint_path(self):
        """
        Validates that the checkpoint path is a valid S3 path.

        Raises:
            ValueError: If the checkpoint path is not a valid S3 path.
        """
        if self.checkpoint_path:
            if not self.checkpoint_path.startswith('s3a://'):
                raise ValueError(f"Checkpoint path must start with 's3a://'. Got: {self.checkpoint_path}")

            # Split into bucket and key
            try:
                path = self.checkpoint_path[6:]  # Remove 's3a://'
                bucket = path.split('/')[0]

                # Basic bucket name validation
                if not bucket or not re.match(r'^[a-z0-9][a-z0-9.-]*[a-z0-9]$', bucket):
                    raise ValueError(f"Invalid S3 bucket name in checkpoint path: {bucket}")

            except Exception as e:
                raise ValueError(f"Invalid S3 path structure: {self.checkpoint_path}. Error: {str(e)}")


    def to_json(self) -> Dict:
        return {
            "deploy-mode": self.deploy_mode or self.config.deploy_mode,
            "spark.kubernetes.driver.memory": self.driver_memory
            or self.config.driver_memory,
            "spark.kubernetes.executor.memory": self.executor_memory
            or self.config.executor_memory,
            "spark.kubernetes.authenticate.driver.serviceAccountName": self.service_account
            or self.config.service_account,
            "spark.hadoop.fs.s3a.aws.credentials.provider": self.aws_credential_provider
            or self.config.aws_credential_provider,
            "spark.kubernetes.container.image": self.container_image
            or self.config.container_image,
            "spark.kubernetes.container.image.pullSecrets": self.container_image_pull_secrets
            or self.config.container_image_pull_secrets,
            "spark.kubernetes.file.upload.path": self.file_upload_path
            or self.config.file_upload_path,
            "spark.kubernetes.namespace": self.name_space or self.config.name_space,
            "num-executors": self.num_executors or self.config.num_executors,
            "checkpoint_path": self.checkpoint_path or self.config.checkpoint_path,
            "spark.executor.cores": self.executor_cores or self.config.executor_cores,
            "spark.driver.cores": self.driver_cores or self.config.driver_cores,
            "spark.kubernetes.driver.limit.cores": self.driver_limit_cores
            or self.config.driver_limit_cores,
            "spark.kubernetes.executor.limit.cores": self.executor_limit_cores
            or self.config.executor_limit_cores,
            "spark.executor.instances": self.executor_instances
            or self.config.executor_instances,
        }

class StreamingProcessingEngineConfigs(ProcessingEngineConfigs):
    """
    Processing Engine Configs for Streaming Features.
    """
    default_config_path = "./gru/config/features/default_processing_engine_configs_streaming.yaml"

class BatchProcessingEngineConfigs(ProcessingEngineConfigs):
    """
    Processing Engine Configs for Raw (Batch) Features.
    """
    default_config_path = "./gru/config/features/default_processing_engine_configs_batch.yaml"



from abc import ABC, abstractmethod
from typing import Optional


class LLMClient(ABC):
    @abstractmethod
    async def generate(self, system_prompt: str, user_prompt: str) -> Optional[str]:
        pass



from typing import Dict, List, Optional
from gru import DEFAULT_CANSO_API_CONFIG
from gru.schema.api_request_handler import APIRequestHandler
from gru.schema.api_response_handler import APIResponseHandler
from gru.utils.config_reader import ConfigReader
from gru.risk.workflows.models.workflows import (
    CreateWorkflowWithRuleRequest, 
    DeployWorkflowRequest, 
    DeployWorkflowResponse, 
    UpdateWorkflowStatusRequest, 
    UpdateWorkflowStatusResponse, 
    Workflow, 
    WorkflowDetails, 
    WorkflowListQueryParams,
    HPAConfig,
)

def get_workflows(auth_token: str, is_active: Optional[bool] = None) -> List[Workflow]:
    """
    Get all workflows via the Risk Workflow Management API.
    
    Args:
        auth_token: Authentication token
        is_active: Optional filter for active/inactive workflows
        
    Returns:
        List of Workflow objects
        
    Raises:
        ApiError: If the API request fails
    """
    # Create query params if needed
    params = {}

    if is_active is not None:
        params["is_active"] = is_active
        
    query_params = WorkflowListQueryParams(**params)
    
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint)
    
    request_handler.set_params(query_params.model_dump(exclude_none=True))

    response = request_handler.send_request(
        "get", 
        configs.workflow_list_endpoint,
    )
    

    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    workflows_data = response_handler.get_success_data()
    return [Workflow.model_validate(wf) for wf in workflows_data]

def get_workflow(auth_token: str, workflow_name: str) -> WorkflowDetails:
    """
    Get detailed workflow information by name from the Risk Workflow Management API.
    
    Args:
        auth_token: Authentication token
        workflow_name: Name of the workflow to retrieve
        
    Returns:
        WorkflowDetails object with complete workflow information
        
    Raises:
        ApiError: If the API request fails
    """
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint)

    endpoint = configs.workflow_get_endpoint.replace('{workflow-name}', workflow_name)
    
    response = request_handler.send_request("get", endpoint)
    
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    workflow_data = response_handler.get_success_data()
    return WorkflowDetails.model_validate(workflow_data)

def create_workflow(auth_token: str, request: CreateWorkflowWithRuleRequest) -> WorkflowDetails:
    """
    Create a new workflow via the Risk Workflow Management API.
    
    Args:
        auth_token: Authentication token
        request: CreateWorkflowRequest containing workflow configuration
        
    Returns:
        WorkflowDetails of created workflow
        
    Raises:
        ApiError: If the API request fails
    """
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint)

    response = request_handler.send_request(
        "post",
        configs.workflow_create_endpoint,
        json_data=request.model_dump(exclude_none=True)
    )
    
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    workflow_data = response_handler.get_success_data()
    return WorkflowDetails.model_validate(workflow_data)

def update_workflow_status(auth_token: str, workflow_name: str, status: str) -> UpdateWorkflowStatusResponse:
    """
    Update workflow status via the Risk Workflow Management API.
    
    Args:
        auth_token: Authentication token
        workflow_name: Name of the workflow to update
        status: New status string (will be validated)
        
    Returns:
        Updated Workflow object
        
    Raises:
        ApiError: If the API request fails
        ValueError: If status is invalid
    """
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint)

    endpoint = configs.workflow_update_status_endpoint.replace('{workflow-name}', workflow_name)
    
    request_data = UpdateWorkflowStatusRequest.from_str(status)
    
    response = request_handler.send_request(
        "patch",
        endpoint,
        json_data=request_data.model_dump()
    )
    
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    workflow_data = response_handler.get_success_data()

    return UpdateWorkflowStatusResponse.model_validate(workflow_data)

def deploy_workflow(
    auth_token: str, 
    workflow_name: str,
    cluster_name: str,
    environment: str,
    namespace: str,
    env_vars: Optional[Dict[str, str]] = None,
    hpa_config: Optional[HPAConfig] = None,
) -> DeployWorkflowResponse:
    """
    Deploy a workflow to specified environment via the Risk Workflow Management API.
    
    Args:
        auth_token: Authentication token
        workflow_name: Name of the workflow to deploy
        cluster_name: Name of the target cluster
        environment: Target environment (STAGING/PRODEXPERIMENT/LIVE)
        namespace: Kubernetes namespace
        env_vars: Optional environment variables for deployment
        hpa_config: Optional HPAConfig object for horizontal pod autoscaling
        
    Returns:
        Deployment response object
        
    Raises:
        ApiError: If the API request fails
        ValueError: If environment is invalid
    """
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint)
    
    endpoint = configs.workflow_deploy_endpoint.replace('{workflow-name}', workflow_name)
    
    allowed_environments = {'BACKTESTING', 'STAGING', 'PRODEXPERIMENT', 'LIVE'}
    if environment.upper() not in allowed_environments:
        raise ValueError(f"Invalid environment: {environment}. Must be one of {allowed_environments}.")

    request_data = DeployWorkflowRequest(
        cluster_name=cluster_name,
        environment=environment.upper(), 
        namespace=namespace,
        env_vars=env_vars or {},
        hpa_configs=hpa_config
    )
    
    response = request_handler.send_request(
        "post",
        endpoint,
        json_data=request_data.model_dump(exclude_none=True)
    )
    
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    deployment_data = response_handler.get_success_data()
    return DeployWorkflowResponse.model_validate(deployment_data)




"""
    Enum-like class defining entity types supported by our Platform

    EntityType is an enumeration class that defines various entity types supported by the platform. 
    These entity types are used to categorize and manage different types of objects and components in the system.
"""

from enum import Enum, unique


@unique
class EntityType(Enum):
    """
    Enum-like class defining entity types supported by our Platform
    """

    DATA_SOURCE = "data_source"
    DATA_SINK = "data_sink"
    RAW_FEATURE = "raw_feature"
    DERIVED_FEATURE = "derived_batch_feature"
    STREAMING_FEATURE = "streaming_feature"
    PRE_PROCESSING_TRANSFORM = "pre_processing_transform"
    TRAINING_DATA = "training_data"
    INFRASTRUCTURE = "infra_details"

    @classmethod
    def get_key(cls, value):
        """
        Get the key (name) of a data type based on its value.

        Args:
            value (str): The value of the data type.

        Returns:
            str: The key (name) of the data type, or None if not found.
        """
        for item in cls:
            if item.value == value:
                return item.name
        return None



from abc import ABC, abstractmethod
from gru.agents.tools.core.code_generator.models import QueryIntent


class QueryAnalyzer(ABC):
    def __init__(self, llm_client):
        self.llm_client = llm_client

    @abstractmethod
    def analyze_query(self, query: str) -> QueryIntent:
        pass



from typing import Any, Dict, Optional
from pydantic import BaseModel

class MemoryStoreRequest(BaseModel):
    collection_name: str
    text: str
    data: Optional[Dict[str, Any]] = None

class MemoryRetrieveParams(BaseModel):
    collection_name: str
    query: str
    top_k: Optional[int] = 5

class MemoryUpdateRequest(BaseModel):
    collection_name: str
    memory_id: str
    text: str
    data: Optional[Dict[str, Any]] = None

class MemoryDeleteRequest(BaseModel):
    collection_name: str
    memory_id: str

class MemoryListCollectionsRequest(BaseModel):
    pass

class MemoryGetCollectionInfoRequest(BaseModel):
    collection_name: str

class MemoryResponse(BaseModel):
    status: str
    message: str
    data: Optional[Dict[str, Any]] = None


from contextlib import asynccontextmanager
import json
import os
from typing import Optional
from fastapi import BackgroundTasks, Depends, FastAPI, Response
import uvicorn
from gru.agents.checkpoint.task_results import TaskResultsRepository, TaskStatus
from gru.agents.framework_wrappers import AgentWorkflow
from gru.agents.schemas import AgentInvokeRequest, AgentInvokeResponse
from gru.agents.schemas.schemas import AgentConversationRequest, TaskCompleteRequest
import logging
from gru.agents.utils.logging import get_log_fields

from gru.agents.schemas.memory import (
    MemoryStoreRequest,
    MemoryRetrieveParams,
    MemoryUpdateRequest,
    MemoryDeleteRequest,
    MemoryResponse
)
from gru.agents.framework_wrappers.memory import CansoMemory

agent_name = os.getenv("AGENT_NAME")

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

@asynccontextmanager
async def lifespan(app: FastAPI):

    workflow: AgentWorkflow = app.state.workflow
    await workflow.setup()
    app.state.task_results_repo = TaskResultsRepository()
    await app.state.task_results_repo.setup()
    yield


api = FastAPI(lifespan=lifespan)

def get_memory() -> CansoMemory:
    return api.state.memory

async def invoke_workflow(request: AgentInvokeRequest):
    try:
        workflow: AgentWorkflow = api.state.workflow
        output = await workflow.invoke(request)
        # Todo: Save output to DB table
        print(output)
    except Exception as e:
        logger.error(
            f"AI agent: invoke api failed - {e}",
            extra=get_log_fields(correlation_id=request.prompt_id),
        )


async def resume_workflow(request: TaskCompleteRequest):
    try:
        workflow: AgentWorkflow = api.state.workflow
        output = await workflow.resume(request)
        # Todo: Save output to DB table
        print(output)
    except Exception as e:
        logger.error(
            f"AI agent: resume workflow failed: {e}",
            extra=get_log_fields(correlation_id=request.prompt_id),
        )


async def update_task_result(request: TaskCompleteRequest):
    try:
        task_results_repo: TaskResultsRepository = api.state.task_results_repo
        await task_results_repo.update(
            agent_name,
            request.prompt_id,
            request.task_type,
            request.tool_call_id,
            TaskStatus.COMPLETED,
            json.dumps(request.result),
        )
    except Exception as e:
        logger.error(
            f"AI agent: Error while upddating task result - {e}",
            extra=get_log_fields(correlation_id=request.prompt_id),
        )


@api.post("/invoke")
async def invoke(
    request: AgentInvokeRequest, background_tasks: BackgroundTasks
) -> AgentInvokeResponse:
    background_tasks.add_task(invoke_workflow, request)
    return AgentInvokeResponse(prompt_id=request.prompt_id)


@api.post("/converse")
async def converse(request: AgentConversationRequest):
    try:
        workflow: AgentWorkflow = api.state.workflow
        return await workflow.converse(request)
    except Exception as e:
        logger.error(
            f"AI agent converse api failed: {e}",
            extra=get_log_fields(correlation_id=request.conversation_id),
        )
        raise e


@api.post("/task-complete")
async def task_complete(
    request: TaskCompleteRequest, background_tasks: BackgroundTasks
):
    background_tasks.add_task(resume_workflow, request)
    return Response(status_code=200)


@api.post("/save-task-result")
async def save_task_result(
    request: TaskCompleteRequest, background_tasks: BackgroundTasks
):
    background_tasks.add_task(update_task_result, request)
    return Response(status_code=200)


@api.post("/memory", response_model=MemoryResponse)
async def store_memory(request: MemoryStoreRequest, memory: CansoMemory = Depends(get_memory)
):

    result = await memory.store(
        data={
            "text": request.text,
            "data": request.data
        },
        collection_name=request.collection_name
    )

    if result.status == "error":
        return MemoryResponse(
            status=result.status,
            message=result.message,
            data={"memory_id": None}
        )

    return MemoryResponse(
        status=result.status,
        message=result.message,
        data={"memory_id": result.ids[0]}
    )

@api.get("/memory", response_model=MemoryResponse)
async def retrieve_memory(params: MemoryRetrieveParams = Depends(), memory: CansoMemory = Depends(get_memory)):
    try:
        results = await memory.retrieve(
            query=params.query,
            collection_name=params.collection_name,
            top_k=params.top_k or 5
        )

        return MemoryResponse(
            status="success",
            message="Retrieved matching documents",
            data={"results": results.results}
        )
    except Exception as e:
        print(f"Error in retrieve endpoint: {str(e)}")
        return MemoryResponse(
            status="error",
            message=f"Error retrieving documents: {str(e)}",
            data={"results": []}
        )

@api.patch("/memory", response_model=MemoryResponse)
async def update_memory(request: MemoryUpdateRequest, memory: CansoMemory = Depends(get_memory)):
    try:
        result = await memory.update(
            doc_id=request.memory_id,
            data={
                "text": request.text,
                "data": request.data
            },
            collection_name=request.collection_name
        )

        if result.status == "error":
            return MemoryResponse(
                status="error",
                message=result.message,
                data={}
            )

        return MemoryResponse(
            status="success",
            message="Document updated successfully",
            data={"count": result.updated_count}
        )
    except Exception as e:
        return MemoryResponse(
            status="error",
            message=f"Error updating document: {str(e)}",
            data=None
        )

@api.delete("/memory", response_model=MemoryResponse)
async def delete_memory(request: MemoryDeleteRequest, memory: CansoMemory = Depends(get_memory)):

    result = await memory.delete(
        doc_id=request.memory_id,
        collection_name=request.collection_name
    )

    return MemoryResponse(
        status=result.status,
        message=result.message,
    )

@api.get("/memory/collections", response_model=MemoryResponse)
async def list_collections(memory: CansoMemory = Depends(get_memory)):
    collections = await memory.vdb_client.list_collections()
    
    return MemoryResponse(
        status=collections.status,
        message=collections.message,
        data={"collections": collections.collections}
    )

@api.get("/memory/collections/{collection_name}", response_model=MemoryResponse)
async def get_collection_info(collection_name: str, memory: CansoMemory = Depends(get_memory)):
        
    info = await memory.vdb_client.get_collection_info(collection_name)
    
    return MemoryResponse(
        status=info.status,
        message=info.message,
        data={"collection_info": info.collection_info}
    )


class App:


    def __init__(self, workflow: AgentWorkflow, memory: Optional[CansoMemory] = None):

        api.state.workflow = workflow
        if memory is not None:
            api.state.memory = memory

    def run(self):
        uvicorn.run(api, host="0.0.0.0", port=8080)



from typing import Optional, Type
from langchain_core.tools import BaseTool
from pydantic import BaseModel, Field
from langchain_core.callbacks import CallbackManagerForToolRun
from gru.agents.tools.core.code_loader.github import GitHubLoader


class GitRepoContentToolInput(BaseModel):
    repository: str = Field(
        description="repository of which the contents are to be retrieved"
    )


class GitRepoContentRetriever(BaseTool):
    name: str = "get_git_repo_contents"
    description: str = (
        "Use this to get file paths of all the files in a github repository"
    )
    args_schema: Type[BaseModel] = GitRepoContentToolInput
    return_direct: bool = True

    class Config:
        extra = "allow"

    def __init__(self, token: str):
        super().__init__(token=token)
        self.token = token
        self.code_loader = GitHubLoader(token)

    def _run(self, *args, **kwargs):
        return super()._run(*args, **kwargs)

    async def _arun(
        self,
        repository: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ):
    
        result = self.code_loader.list_contents(repository)
        return result



"""
Sinks are an abstraction that define where outputs from various jobs should be persisted
"""

from abc import abstractmethod
from typing import Any, Dict, List, Optional, Union
import json
from datetime import datetime
from gru.schema.registrable import Registrable
from gru.sources.file_types import CSVType, ParquetType
from gru.utils.entity_type import EntityType


class Sink(Registrable):
    """
    Base class for all Sinks
    """

    def __init__(
        self,
        name: str,
        family: str,
        physical_uri: Dict[str, Any],
        type: str = None,
        file_type: Union[CSVType, ParquetType] = None,
        file_type_properties: Optional[Dict[str, Any]] = None,
        cloud_provider: str = None,
        description: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = {},
        owners: Optional[List] = None,
        created_at=datetime.now(),
        # Maybe Default value comes from a separate owners class?
    ):
        """

        Arguments
        ---------
        name: Unique indentifier for the Sink
        family: Unique indentifier for the Sink Family e.g. ObjectStorage, Kafka, MySQL etc.
        physical_uri: All properties that identify the sink resource
        type: Unique name of a sink within a cloud provider
        file_type (optional) - If not None, it specifies the format of the data, e.g. csv, json, parquet, etc.
        file_type_properties (optional) Properties of the file_type.
        cloud_provider - The cloud provider e.g. aws, gcp, etc
        description (optional) - A human-readable description.
        """
        self.name = name
        self.family = family
        self.physical_uri = physical_uri
        self.entity_type = EntityType.DATA_SINK.value
        self.type = type
        self.file_type = file_type
        self.file_type_properties = file_type_properties
        self.cloud_provider = cloud_provider
        self.description = description
        self.metadata = metadata
        self.owners = owners
        self.created_at = created_at

    @abstractmethod
    def to_json(self) -> Dict:
        """
        Returns a json-formatted object representing this entity.
        """
        pass

    def __str__(self) -> str:
        return json.dumps(self.to_json(), indent=2, sort_keys=True)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.to_json() == other.to_json()
    
    def to_register_json(self) -> Dict:
        return self.to_json()



from typing import Dict, Any


class CustomFeatureLogic:
    """
    Creates CustomFeatureLogic object for the feature logic provided from custom UDF.

    Attributes:
        custom_class_name (str): Class name from custom file.
        file_path (str): File path containing custom UDF script.
        is_feature_logic_custom (bool): Flag to decide weather the logic is custom or predefined, Default value is True.

    Examples
    --------
    >>>    custom_class_name="CreditScoreProcessor",
    ...    custom_file_path="/opt/spark/work-dir/bob/src/v2/external_udfs/features/custom_udf.py",
    ...    custom_docker_image="private-repo/custom-image:v1.0",
    ...    custom_feature_args={
    ...        "timestamp_column": "timestamp",
    ...        "groupby_keys": ["id", "timestamp"]
    ...    },
    ...    mode="spark_sql"
    """

    def __init__(
        self,
        class_name: str = "",
        file_path: str = "",
        docker_image: str = "",
        feature_args: Dict[str, Any] = {},
        mode: str = "",
    ):
        self.custom_class_name = class_name
        self.custom_file_path = file_path
        self.function_type = "custom_udf"
        self.custom_docker_image = docker_image
        self.custom_feature_args = feature_args
        self.mode = mode

    def to_json(self) -> Dict:
        """
        Returns:
            Dict: A dictionary representing the feature attributes in JSON format.
        """
        return {
            "custom_class_name": self.custom_class_name,
            "custom_file_path": self.custom_file_path,
            "custom_docker_image": self.custom_docker_image,
            "function_type": self.function_type,
            "mode": self.mode,
            **self.custom_feature_args,
        }



"""
Abstract base class for web framework adapters.

This class defines a set of methods that act as a bridge between the generic server and the specific features of a web framework. 
It provides a framework for making predictions, processing requests, formatting and sending responses, loading user BYOC (Bring Your Own Code) code, and managing assets and server startup.

Methods:
    - __init__: Initialize the WebFrameworkAdapter instance with a configuration reader.
    - predict: Abstract method for making predictions.
    - process_request: Abstract method for processing incoming requests.
    - format_response: Abstract method for formatting and sending responses.
    - load_application: Abstract method for loading user BYOC code.
"""

from abc import ABC, abstractmethod
from gru.ml_serving.ml_serving_config_reader import MLServingConfigReader

class WebFrameworkAdapter(ABC):

    def __init__(self, config: MLServingConfigReader):
        """
        Initialize the WebFrameworkAdapter instance with a configuration reader.

        Parameters:
            - config: MLServingConfigReader instance for reading configuration settings.
        """
        self.config = config

    @abstractmethod
    def predict(self, input_data):
        """
        Make predictions based on input data.

        Parameters:
            - input_data: Input data for making predictions.

        Returns:
            Any: Predicted results.
        """
        pass

    @abstractmethod
    def process_request(self, request):
        """
        Process incoming requests.

        Parameters:
            - request: Request object received from the web framework.

        Returns:
            dict: Processed request data.
        """
        pass

    @abstractmethod
    def format_response(self, response_data):
        """
        Format and send responses.

        Parameters:
            - response_data: Data to be included in the response.

        Returns:
            Any: Formatted response.
        """
        pass

    @abstractmethod
    def load_application(self, code_path):
        """
        Load user BYOC (Bring Your Own Code) from the specified path.

        Parameters:
            - code_path: Path to the user code.

        Returns:
            Any: Loaded user code.
        """
        pass



"""
Aggregations that can be used to define Features.
Functions for Derived Features
- add
- subtract
- multiply
- safe_divide
- binarize
- bucketize

`Spark's Bucketizer Feature Transformer <https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.ml.feature.Bucketizer.html>`_
"""
import warnings
from typing import Union, List, Dict, Any

# TODO: [RJS 2024-06-20] - Modify the docstring for aggregation functions used for derived feature.


class WindowAggregation:
    """
    WindowAggregation is used for calculating aggregations on given features for a fixed window in streaming data.

    This class supports various aggregation functions and allows for customization of the window parameters.

    Parameters
    ----------
    function : str
        Aggregation function to be performed on the given feature (e.g., "sum", "avg", "max", "min", "count").
    partition_by : str, optional
        Column(s) to partition the data by before applying the window function.
    order_by : str, optional
        Column(s) to order the data by within each partition.
    window_duration : str, optional
        The duration of the window (e.g., "5 seconds", "1 minute", "1 hour").
    watermark_delay_threshold : str, optional
        The maximum time to wait for late data. Events arriving later than this will be dropped.

    Attributes
    ----------
    transform_args : dict
        A dictionary containing all the parameters for the window aggregation, including the function type.

    Examples
    --------
    >>> window_agg = WindowAggregation(
    ...     function="sum",
    ...     partition_by="user_id",
    ...     order_by="timestamp",
    ...     window_duration="5 minutes",
    ...     watermark_delay_threshold="10 seconds"
    ... )

    """
    def __init__(
        self,
        function: str,
        partition_by: str = None,
        order_by: str = None,
        window_duration: str = None,
        watermark_delay_threshold: str = None
    ):
        self.function = function
        self.partition_by = partition_by
        self.order_by = order_by
        self.window_duration = window_duration
        self.watermark_delay_threshold = watermark_delay_threshold
        self.transform_args = {
            "function": self.function,
            "partition_by": self.partition_by,
            "order_by": self.order_by,
            "window_duration": self.window_duration,
            "watermark_delay_threshold": self.watermark_delay_threshold,
            "function_type": "window_aggregation",
        }

class SessionWindowAggregation:
    """
    SessionWindowAggregation is used for calculating aggregations over dynamic session windows in streaming data.
    Sessions are defined by periods of activity separated by a gap of inactivity.

    This class supports various aggregation functions and allows for customization of session parameters.

    Parameters
    ----------
    function : str
        Aggregation function to be performed on the given feature (e.g., "sum", "avg", "max", "min", "count").
    partition_by : str, optional
        Column(s) to partition the data by before applying the session window function.
    order_by : str, optional
        Column(s) to order the data by within each partition.
    gap_duration : str, optional
        The duration of inactivity that defines the end of a session (e.g., "30 minutes", "1 hour", "2 hours").
        A new session starts when the gap between consecutive events exceeds this duration.
    watermark_delay_threshold : str, optional
        The maximum time to wait for late data. Events arriving later than this will be dropped.

    Attributes
    ----------
    transform_args : dict
        A dictionary containing all the parameters for the session window aggregation, including the function type.

    Examples
    --------
    >>> session_window_agg = SessionWindowAggregation(
    ...     function="count",
    ...     partition_by="user_id",
    ...     order_by="event_timestamp",
    ...     gap_duration="30 minutes",
    ...     watermark_delay_threshold="1 hour"
    ... )

    Notes
    -----
    - Unlike fixed windows, session windows are dynamic and their duration depends on the event pattern.
    - The gap_duration parameter defines the maximum allowed time between events within the same session.
    - Sessions are particularly useful for analyzing user behavior patterns, website visits, and IoT device activity.
    """
    def __init__(
        self,
        function: str,
        partition_by: str = None,
        order_by: str = None,
        gap_duration: str = None,
        watermark_delay_threshold: str = None
    ):
        self.function = function
        self.partition_by = partition_by
        self.order_by = order_by
        self.gap_duration = gap_duration
        self.watermark_delay_threshold = watermark_delay_threshold
        self.transform_args = {
            "function": self.function,
            "partition_by": self.partition_by,
            "order_by": self.order_by,
            "gap_duration": self.gap_duration,
            "watermark_delay_threshold": self.watermark_delay_threshold,
            "function_type": "session_window_aggregation",
        }

class SlidingWindowAggregation:
    """
    SlidingWindowAggregation is used for calculating aggregations on given features using a sliding window approach
    in both batch and streaming data scenarios.

    This class supports various aggregation functions and allows for customization of the window parameters,
    including range-based windows, row-based windows, and streaming windows.

    Parameters
    ----------
    function : str
        Aggregation function to be performed on the given feature (e.g., "sum", "avg", "max", "min", "count").
    partition_by : str
        Column(s) to partition the data by before applying the window function.
    rangeBetween : Dict, optional
        A dictionary specifying the range-based window. It should contain 'frame_start' and 'frame_end' keys.
    rowsBetween : Dict, optional
        A dictionary specifying the row-based window. It should contain 'frame_start' and 'frame_end' keys.
    window_period : str, optional
        The interval at which the window slides forward (for streaming windows).
    window_duration : str, optional
        The duration of the window (for streaming windows).
    watermark_delay_threshold : str, optional
        The maximum time to wait for late data in streaming scenarios. Events arriving later than this will be dropped.
    order_by : List[str], optional
        Column(s) to order the data by within each partition.

    Attributes
    ----------
    transform_args : dict
        A dictionary containing all the parameters for the sliding window aggregation, including the function type.

    Examples
    --------
     >>> streaming_window = SlidingWindowAggregation(
    ...     function="count",
    ...     partition_by="category",
    ...     order_by=["event_time"],
    ...     window_period="1 minute",
    ...     window_duration="5 minutes",
    ...     watermark_delay_threshold="10 seconds"
    ... )
    """

    def __init__(
        self,
        function: str,
        partition_by: str = "",
        rangeBetween: Dict = None,
        rowsBetween: Dict = None,
        window_period: str = None,
        window_duration: str = None,
        watermark_delay_threshold: str = "",
        order_by: List[str] = [],
    ):
        warnings.warn(
            "This is currently a partially-implemented feature. If a data source has different folders/keys (e.g. one key for each day or one key for each hour etc.) this may lead to unexpected and inaccurate results. This is because additional data may need to be read to complete the window that has been specified for some rows. A follow-up release will allow the specification of an optional cushion argument, so that each feature job can read additional keys/folders of data to accommodate the window frame provided for all rows."
        )

        self.function = function
        self.partition_by = partition_by
        self.order_by = order_by
        self.rangeBetween = rangeBetween
        self.rowsBetween = rowsBetween
        self.window_period = window_period
        self.window_duration = window_duration
        self.watermark_delay_threshold = watermark_delay_threshold
        if self.rangeBetween is not None:
            frame_start_range = self.frame_start_range_val = self.rangeBetween.get(
                "frame_start"
            )
            frame_end_range = self.frame_end_range_val = self.rangeBetween.get(
                "frame_end"
            )
            self.frame_start_range_exists = frame_start_range is not None
            self.frame_end_range_exists = frame_end_range is not None
            if self._validate_range():
                self._validate_frame_range_valid_window()

        if rowsBetween is not None:
            self.frame_start_rows = self.frame_start_rows_val = rowsBetween.get(
                "frame_start"
            )
            self.frame_end_rows = self.frame_end_rows_val = rowsBetween.get("frame_end")
            self.frame_start_rows_exists = self.frame_start_rows is not None
            self.frame_end_rows_exists = self.frame_end_rows is not None
            if self._validate_rows():
                self._validate_frame_rows_valid_window()

        if window_period != None and window_duration != None:
            self.streaming_slidingwindow_args()

    def _validate_order_by(self, order_by: List[str]) -> bool:
        """
        Validate that the specified order column is not empty.

        Args:
            order_by (List[str]): The name of the column to use for ordering.

        Raises:
            ValueError: If the order_by is empty.

        Returns:
            bool

        """
        if len(order_by) == 0:
            raise ValueError("order_by should not be empty")
        return True

    def _validate_range(self) -> bool:
        """
        Validate that the  start range value  and end range value should not be None.
        Raises:
            ValueError: If frame_start_range and frame_end_range is None .
        Returns:
            bool
        """
        if not self.frame_start_range_exists:
            raise ValueError(
                "Please enter the value of frame_start_range. It should not be null"
            )
        elif not self.frame_end_range_exists:
            raise ValueError(
                "Please enter the value of frame_end_range. It should not be null"
            )
        return (
            self.frame_start_range_exists
            and self.frame_end_range_exists
            and self._validate_order_by(self.order_by)
        )

    def _validate_frame_range_valid_window(self) -> Dict:
        """
        Validate that the  start range value  should be less than equal to  end range value .
        Raises:
            ValueError: If frame_start_range is greater frame_end_range is None .
        Returns:
            dict
        """
        if (
            self.frame_start_range_exists
            and self.frame_end_range_exists
            and (self.frame_start_range_val <= self.frame_end_range_val)
        ):
            self.transform_args = {
                "function": self.function,
                "partition_by": self.partition_by,
                "order_by": self.order_by,
                "rangeBetween": self.rangeBetween,
                "function_type": "sliding_window_aggregation",
            }
        else:
            raise ValueError(
                "The start value of the range should be smaller than or equal to the end value of the range."
            )

    def streaming_slidingwindow_args(self) -> Dict:
        self.transform_args = {
            "function": self.function,
            "partition_by": self.partition_by,
            "order_by": self.order_by,
            "window_period": self.window_period,
            "window_duration": self.window_duration,
            "watermark_delay_threshold": self.watermark_delay_threshold,
            "function_type": "sliding_window_aggregation",
        }

    def _validate_rows(self) -> bool:
        """
        Validate that the  start rows value  and end rows value should not be None.
        Raises:
            ValueError: If frame_start_rows and frame_end_rows is None .
        Returns:
            bool
        """
        if not self.frame_start_rows_exists:
            raise ValueError(
                " Please enter the value of frame_start_rows. It should not be null"
            )
        elif not self.frame_end_rows_exists:
            raise ValueError(
                "Please enter the value of frame_end_rows. It should not be null"
            )
        return (
            self.frame_start_rows_exists
            and self.frame_end_rows_exists
            and self._validate_order_by(self.order_by)
        )

    def _validate_frame_rows_valid_window(self) -> Dict:
        """
        Validate that the  start rows value  should be less than equal to  end rows value .
        Raises:
            ValueError: If  value of frame_start_rows is greater than  frame_end_rows.
        Returns:
            dict
        """
        if (
            self.frame_start_rows_exists
            and self.frame_end_rows_exists
            and (self.frame_start_rows_val <= self.frame_end_rows_val)
        ):
            self.transform_args = {
                "function": self.function,
                "partition_by": self.partition_by,
                "order_by": self.order_by,
                "rowsBetween": self.rowsBetween,
                "function_type": "sliding_window_aggregation",
            }
        else:
            raise ValueError(
                "The start value of the rows should be smaller than or equal to the end value of the rows."
            )


def add(feature_list: List[str], groupby_keys: List[str]) -> Dict[str, Any]:
    """
    Aggregation function to create a derived feature by adding multiple raw/derived features

    Parameters
    ----------
    *args: accepting variable number of feature names

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "add",
            "feature_list": ['feature_1', 'feature_2', ...]
        }

    Notes
    -----
    Only feature names will be accepted as arguments

    Examples
    --------

    """

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "add",
        "feature_list": feature_list,
        "groupby_keys": groupby_keys,
    }
    return obj


def subtract(feature_list: List[str], groupby_keys: List[str]) -> Dict[str, Any]:
    """
    Aggregation function to create a derived feature by subtracting feature1 from feature2

    Parameters
    ----------
    feature1: string
        name of the feature to be subtracted
    feature2: string
        name of the feature from which we'll subtract

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "subtract",
            "feature_list": ['feature_1', 'feature_2']
        }

    Notes
    -----
    Only feature names will be accepted as arguments

    Examples
    --------

    """

    if len(feature_list) != 2:
        raise ValueError(
            "feature_list must contain exactly two elements: ['feature1', 'feature2']"
        )

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "subtract",
        "feature_list": feature_list,
        "groupby_keys": groupby_keys,
    }
    return obj


def multiply(feature_list: List[str], groupby_keys: List[str]) -> Dict[str, Any]:
    """
    Aggregation function to create a derived feature by multiplying multiple raw/derived features

    Parameters
    ----------
    *args: accepting variable number of feature names

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "multiply",
            "feature_list": ['feature_1', 'feature_2', ...]
        }

    Notes
    -----
    Only feature names will be accepted as arguments

    Examples
    --------

    """
    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "multiply",
        "feature_list": feature_list,
        "groupby_keys": groupby_keys,
    }
    return obj


def safe_divide(feature_list: List[str], groupby_keys: List[str]) -> Dict[str, Any]:
    """
    Aggregation function to create a derived feature by safe dividing feature1 from feature2

    Parameters
    ----------
    feature1: string
        name of the feature to be divided
    feature2: string
        name of the feature from which we'll divide

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "safe_divide",
            "feature_list": ['feature_1', 'feature_2']
        }

    Notes
    -----
    Only feature names will be accepted as arguments

    Examples
    --------

    """
    if len(feature_list) != 2:
        raise ValueError(
            "feature_list must contain exactly two elements: ['feature1', 'feature2']"
        )

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "safe_divide",
        "feature_list": feature_list,
        "groupby_keys": groupby_keys,
    }
    return obj


def binarize(
    feature: str,
    threshold: Union[int, float],
    groupby_keys: List,
    if_true: Union[int, float, bool, str] = 1,
    if_false: Union[int, float, bool, str] = 0,
) -> Dict[str, Any]:
    """
    Performs binarization by thresholding numerical features to binary features using Spark's Binarizer Feature Transformer

    Parameters
    ----------
    feature: str
        feature name to binarize
    threshold: Union[int, float]
        threshold value can either be an int or float. Feature will be binarized based on the threshold value
    if_true: Union[int, float, bool, str]
        For feature values greater than threshold, if_true value will be set to the binarize feature. Default set to 1
    if_false: Union[int, float, bool, str]
        For feature values less than threshold, if_false value will be set to the binarize feature. Default set to 0

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "binarize",
            "feature_list": [feature],
            feature_transform_args : {
                "labels": [if_true, if_false],
                "threshold": threshold,
            }
        }

    Notes
    -----

    Examples
    --------
    >>> create_derived_feature = CreateDerivedFeature(
    ...     feature_name="player_count_test",
    ...     feature_description="Total players",
    ...     feature_data_type="FLOAT",
    ...     owners=["all-ds@company.com", "temp"],
    ...     schedule="*/7 * * * *",
    ...     entity=["player_count"],
    ...     online=False,
    ...     offline=True,
    ...     transform=src.modules.aggregations.binarize(
    ...         feature = "raw_feature_test",
    ...         threshold = 500,
    ...         if_true = "less",
    ...         if_false = "more"
    ...     ),
    ... )
    """

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "binarize",
        "feature_list": [feature],
        "feature_transform_args": {
            "labels": [if_true, if_false],
            "threshold": threshold,
        },
        "groupby_keys": groupby_keys,
    }
    return obj


def bucketize(
    feature: str, groupby_keys: List, splits: List[Union[int, float]], labels: List[str]
) -> Dict[str, Any]:
    """
    Transforms a column of continuous features to a column of feature buckets, where the buckets are specified by users using [Spark's Bucketizer Feature Transformer](https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.ml.feature.Bucketizer.html)

    Parameters
    ----------
    feature: str
        name of the feature to be bucketized
    splits: List[Union[int, float]]
        list of splits based on which a continuous feature will be bucketized
    labels: List[str]
        list of labels that will be assigend as per the splits

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "bucketize",
            "feature_list": [feature],
            "feature_transform_args" : {
                "labels": labels,
                "splits": splits,
            }
        }

    Example
    -------
    >>> create_derived_feature = CreateDerivedFeature(
    ...     feature_name="player_count_test",
    ...     feature_description="Total players",
    ...     feature_data_type="FLOAT",
    ...     owners=["all-ds@company.com", "temp"],
    ...     schedule="*/7 * * * *",
    ...     entity=["player_count"],
    ...     online=False,
    ...     offline=True,
    ...     transform=src.modules.aggregations.bucketize(
    ...         feature = "raw_feature_test",
    ...         splits = [-float("inf"), 500, 750, float("inf")],
    ...         labels = ["less_than_500", "greater_than_500","greater_than_750"],
    ...     ),
    ... )
    ... _
    ... create_derived_feature.register_feature()
    ... create_derived_feature.deploy_feature()
    ..."""
    # converting to string for out of range values
    splits_str = [str(split) for split in splits]

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "bucketize",
        "feature_list": [feature],
        "feature_transform_args": {
            "labels": labels,
            "splits": splits_str,
        },
        "groupby_keys": groupby_keys,
    }
    return obj


def MIN(feature_list: List[str], groupby_keys: List[str]) -> Dict[str, Any]:
    """
    Aggregation function to create a derived feature by finding the minimum of multiple raw/derived features

    Parameters
    ----------
    *args: accepting variable number of feature names

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "min",
            "feature_list": ['feature_1', 'feature_2', ...]
        }

    Notes
    -----
    Only feature names will be accepted as arguments

    Examples
    --------

    """

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "min",
        "feature_list": feature_list,
        "groupby_keys": groupby_keys,
    }
    return obj


def MAX(feature_list: List[str], groupby_keys: List[str]) -> Dict[str, Any]:
    """
    Aggregation function to create a derived feature by finding the maximum of multiple raw/derived features

    Parameters
    ----------
    *args: accepting variable number of feature names

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "max",
            "feature_list": ['feature_1', 'feature_2', ...]
        }

    Notes
    -----
    Only feature names will be accepted as arguments

    Examples
    --------

    """

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "max",
        "feature_list": feature_list,
        "groupby_keys": groupby_keys,
    }
    return obj


def AVG(feature_list: List[str], groupby_keys: List[str]) -> Dict[str, Any]:
    """
    Aggregation function to create a derived feature by calculating the average of multiple raw/derived features

    Parameters
    ----------
    *args: accepting variable number of feature names

    Returns
    -------
    returns a json object as below:


    >>> {
            "transform": "avg",
            "feature_list": ['feature_1', 'feature_2', ...]
        }

    Notes
    -----
    Only feature names will be accepted as arguments

    Examples
    --------

    """

    if not groupby_keys:
        raise ValueError("groupby_keys must contain at least one key")

    obj = {
        "transform": "avg",
        "feature_list": feature_list,
        "groupby_keys": groupby_keys,
    }
    return obj



from typing import Any, Dict, List, Literal, Optional, Union
from gru import DEFAULT_CANSO_API_CONFIG
from gru.schema.api_request_handler import APIRequestHandler
from gru.schema.api_response_handler import APIResponseHandler
from gru.utils.config_reader import ConfigReader
from gru.risk.workflows.models.rules import CreateRuleRequest, RiskRule, UpdateRuleRequest, UpdateRuleStageRequest, UpdateRuleStatusRequest, WorkflowRulesQueryParams

def get_workflow_rules(
    auth_token: str,
    workflow_name: str,
    rule_name: Optional[str] = None,
    stage: Optional[str] = None,
    status: Optional[str] = None
) -> List[RiskRule]:
    """
    Get rules for a workflow, optionally filtering by rule name, stage, and status.
    
    Args:
        auth_token: Authentication token
        workflow_name: Name of the workflow
        rule_name: Optional specific rule name to fetch
        stage: Optional stage filter
        status: Optional status filter
        
    Returns:
        List of RiskRule objects
        
    Raises:
        ApiError: If the API request fails
    """
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint) #TODO Temp fix
    
    endpoint = configs.workflow_rules_list_endpoint.replace(
        '{workflow-name}', 
        workflow_name
    )
    
    if rule_name:
        endpoint = f"{endpoint}/{rule_name}"
    else:
        params = WorkflowRulesQueryParams(stage=stage, status=status)
        request_handler.set_params(params.model_dump(exclude_none=True))
    
    response = request_handler.send_request("get", endpoint)
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    rules_data = response_handler.get_success_data()

    if isinstance(rules_data, list):
        return [RiskRule.model_validate(rule) for rule in rules_data]
    else:
        return [RiskRule.model_validate(rules_data)]
    

def create_rule(auth_token: str, workflow_name: str, rule_config: Dict[str, Any]) -> RiskRule:
    """
    Create a new rule for a workflow via the Risk Workflow Management API.
    
    Args:
        auth_token: Authentication token
        workflow_name: Name of the workflow to add the rule to
        rule_config: Rule configuration from JSON file
        
    Returns:
        RiskRule: The created rule object
        
    Raises:
        ApiError: If the API request fails
        ValueError: If rule configuration is invalid
    """
    rule_request = CreateRuleRequest.model_validate(rule_config)
    
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint)
    
    endpoint = configs.workflow_rules_create_endpoint.format(**{
        "workflow-name": workflow_name
    })
    
    response = request_handler.send_request(
        "post",
        endpoint,
        json_data=rule_request.model_dump(exclude_none=True)
    )
    
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    rule_data = response_handler.get_success_data()
    return RiskRule.model_validate(rule_data)

def update_workflow_rule(
    auth_token: str,
    workflow_name: str,
    rule_name: str,
    update_type: Literal['definition', 'status', 'stage'],
    request_data: Union[UpdateRuleRequest, UpdateRuleStatusRequest, UpdateRuleStageRequest]
) -> RiskRule:
    """
    Update a workflow rule via the Risk Workflow Management API.
    
    Args:
        auth_token: Authentication token
        workflow_name: Name of the workflow
        rule_name: Name of the rule to update
        update_type: Type of update (definition/status/stage)
        request_data: Request data for the update
        
    Returns:
        Updated RiskRule object
        
    Raises:
        ApiError: If the API request fails
    """
    configs = ConfigReader(DEFAULT_CANSO_API_CONFIG)
    request_handler = APIRequestHandler(auth_token, configs)
    request_handler.set_base_url(configs.risk_mgmt_endpoint) 
    endpoint_key = ''
    if update_type == 'definition':
        endpoint_key = 'update'
    elif update_type == 'status':
        endpoint_key = 'update_status'
    elif update_type == 'stage':
        endpoint_key = 'update_stage'
        
    endpoint = configs.risk_paths['rules'][endpoint_key]
    endpoint = endpoint.replace('{workflow-name}', workflow_name)
    endpoint = endpoint.replace('{rule-name}', rule_name)
    
    method = "put" if update_type == 'definition' else "patch"
    
    response = request_handler.send_request(
        method,
        endpoint,
        json_data=request_data.model_dump(exclude_none=True)
    )
    
    response_handler = APIResponseHandler(response)
    response_handler.check_for_errors()
    
    rule_data = response_handler.get_success_data()
    return RiskRule.model_validate(rule_data)
